@@ -31,6 +31,10 @@ extern char **environ;
 #  undef PIPE_STATES_SOCKET
 #endif
 
+#if defined(__sun) || defined(__sun__) || defined(__SunOS)
+#  define XIO_ANCILLARY_TYPE_SOLARIS 1
+#endif
+
 /*****************************************************************************/
 
 /* substitute some features that might be missing on some platforms */
@@ -94,6 +98,111 @@ typedef int sig_atomic_t;
 #  define SOL_IPV6 IPPROTO_IPV6
 #endif
 
+#define F_uint8_t "%hu"
+#define F_uint8_x "%02hx"
+#define F_int8_t  "%hd"
+
+#ifndef F_uint16_t
+#  if HAVE_BASIC_UINT16_T==0
+#    define F_uint16_t "%hu"
+#    define F_uint16_x "%04hx"
+#  elif HAVE_BASIC_UINT16_T==2
+#    define F_uint16_t "%hu"
+#    define F_uint16_x "%04hx"
+#  elif HAVE_BASIC_UINT16_T==4
+#    define F_uint16_t "%u"
+#    define F_uint16_x "%04x"
+#  elif HAVE_BASIC_UINT16_T==6
+#    define F_uint16_t "%lu"
+#    define F_uint16_x "%04lx"
+#  else
+#    error "HAVE_BASIC_UINT16_T is out of range:" HAVE_BASIC_UINT16_T
+#  endif
+#endif
+
+#ifndef F_uint32_t
+#  if HAVE_BASIC_UINT32_T==0
+#    define F_uint32_t "%hu"
+#    define F_uint32_x "%08hx"
+#  elif HAVE_BASIC_UINT32_T==2
+#    define F_uint32_t "%hu"
+#    define F_uint32_x "%08hx"
+#  elif HAVE_BASIC_UINT32_T==4
+#    define F_uint32_t "%u"
+#    define F_uint32_x "%08x"
+#  elif HAVE_BASIC_UINT32_T==6
+#    define F_uint32_t "%lu"
+#    define F_uint32_x "%08lx"
+#  else
+#    error "HAVE_BASIC_UINT32_T is out of range:" HAVE_BASIC_UINT32_T
+#  endif
+#endif
+
+#ifndef F_uint64_t
+#  if HAVE_BASIC_UINT64_T==0
+#    define F_uint64_t "%hu"
+#    define F_uint64_x "%016hx"
+#  elif HAVE_BASIC_UINT64_T==2
+#    define F_uint64_t "%hu"
+#    define F_uint64_x "%016hx"
+#  elif HAVE_BASIC_UINT64_T==4
+#    define F_uint64_t "%u"
+#    define F_uint64_x "%016x"
+#  elif HAVE_BASIC_UINT64_T==6
+#    define F_uint64_t "%lu"
+#    define F_uint64_x "%016lx"
+#  elif HAVE_BASIC_UINT64_T==8
+#    define F_uint64_t "%llu"
+#    define F_uint64_x "%016llx"
+#  else
+#    error "HAVE_BASIC_UINT64_T is out of range:" HAVE_BASIC_UINT64_T
+#  endif
+#endif
+
+#ifndef F_int16_t
+#  if HAVE_BASIC_INT16_T==0
+#    define F_int16_t "%hd"
+#  elif HAVE_BASIC_INT16_T==1
+#    define F_int16_t "%hd"
+#  elif HAVE_BASIC_INT16_T==3
+#    define F_int16_t "%d"
+#  elif HAVE_BASIC_INT16_T==5
+#    define F_int16_t "%ld"
+#  else
+#    error "HAVE_BASIC_INT16_T is out of range:" HAVE_BASIC_INT16_T
+#  endif
+#endif
+
+#ifndef F_int32_t
+#  if HAVE_BASIC_INT32_T==0
+#    define F_int32_t "%hd"
+#  elif HAVE_BASIC_INT32_T==1
+#    define F_int32_t "%hd"
+#  elif HAVE_BASIC_INT32_T==3
+#    define F_int32_t "%d"
+#  elif HAVE_BASIC_INT32_T==5
+#    define F_int32_t "%ld"
+#  else
+#    error "HAVE_BASIC_INT32_T is out of range:" HAVE_BASIC_INT32_T
+#  endif
+#endif
+
+#ifndef F_int64_t
+#  if HAVE_BASIC_INT64_T==0
+#    define F_int64_t "%hd"
+#  elif HAVE_BASIC_INT64_T==1
+#    define F_int64_t "%hd"
+#  elif HAVE_BASIC_INT64_T==3
+#    define F_int64_t "%d"
+#  elif HAVE_BASIC_INT64_T==5
+#    define F_int64_t "%ld"
+#  elif HAVE_BASIC_INT64_T==7
+#    define F_int64_t "%lld"
+#  else
+#    error "HAVE_BASIC_INT64_T is out of range:" HAVE_BASIC_INT64_T
+#  endif
+#endif
+
 /* all unsigned */
 #if !defined(HAVE_BASIC_SIZE_T) || !HAVE_BASIC_SIZE_T
 #  undef HAVE_BASIC_SIZE_T
@@ -350,6 +459,35 @@ typedef int sig_atomic_t;
 #  endif
 #endif
 
+
+#if _WITH_TERMIOS
+#if !defined(HAVE_BASIC_SPEED_T) || !HAVE_BASIC_SPEED_T
+#  undef HAVE_BASIC_SPEED_T
+#  define HAVE_BASIC_SPEED_T 4
+#endif
+#ifndef F_speed
+#  if HAVE_BASIC_SPEED_T==1
+#define F_speed "%hd"
+#  elif HAVE_BASIC_SPEED_T==2
+#define F_speed "%hu"
+#  elif HAVE_BASIC_SPEED_T==3
+#define F_speed "%""d"
+#  elif HAVE_BASIC_SPEED_T==4
+#define F_speed "%u"
+#  elif HAVE_BASIC_SPEED_T==5
+#define F_speed "%ld"
+#  elif HAVE_BASIC_SPEED_T==6
+#define F_speed "%lu"
+#  elif HAVE_BASIC_SPEED_T==7
+#define F_speed "%Ld"
+#  elif HAVE_BASIC_SPEED_T==8
+#define F_speed "%Lu"
+#  else
+#error "HAVE_BASIC_SPEED_T is out of range:" HAVE_BASIC_SPEED_T
+#  endif
+#endif
+#endif /* _WITH_TERMIOS */
+
 /* all unsigned; default; unsigned long */
 #if !defined(HAVE_TYPEOF_ST_INO) || !HAVE_TYPEOF_ST_INO
 #  undef HAVE_TYPEOF_ST_INO
@@ -597,6 +735,33 @@ typedef int sig_atomic_t;
 #  endif
 #endif
 
+/* default: long */
+#if !defined(HAVE_TYPEOF_STRUCT_TIMESPEC_TV_NSEC) || !HAVE_TYPEOF_STRUCT_TIMESPEC_TV_NSEC
+#  undef HAVE_TYPEOF_STRUCT_TIMESPEC_TV_NSEC
+#  define HAVE_TYPEOF_STRUCT_TIMESPEC_TV_NSEC 5
+#endif
+#ifndef F_tv_nsec
+#  if HAVE_TYPEOF_STRUCT_TIMESPEC_TV_NSEC==1
+#define F_tv_nsec "%09hd"
+#  elif HAVE_TYPEOF_STRUCT_TIMESPEC_TV_NSEC==2
+#define F_tv_nsec "%09hu"
+#  elif HAVE_TYPEOF_STRUCT_TIMESPEC_TV_NSEC==3
+#define F_tv_nsec "%09d"
+#  elif HAVE_TYPEOF_STRUCT_TIMESPEC_TV_NSEC==4
+#define F_tv_nsec "%09u"
+#  elif HAVE_TYPEOF_STRUCT_TIMESPEC_TV_NSEC==5
+#define F_tv_nsec "%09ld"
+#  elif HAVE_TYPEOF_STRUCT_TIMESPEC_TV_NSEC==6
+#define F_tv_nsec "%09lu"
+#  elif HAVE_TYPEOF_STRUCT_TIMESPEC_TV_NSEC==7
+#define F_tv_nsec "%09Ld"
+#  elif HAVE_TYPEOF_STRUCT_TIMESPEC_TV_NSEC==8
+#define F_tv_nsec "%09Lu"
+#  else
+#error "HAVE_TYPEOF_STRUCT_TIMESPEC_TV_NSEC is out of range:" HAVE_TYPEOF_STRUCT_TIMESPEC_TV_NSEC
+#  endif
+#endif
+
 /* default: long */
 #if !defined(HAVE_TYPEOF_RLIM_MAX) || !HAVE_TYPEOF_RLIM_MAX
 #  undef HAVE_TYPEOF_RLIM_MAX
@@ -624,6 +789,10 @@ typedef int sig_atomic_t;
 #  endif
 #endif
 
+/* sigset_t printing - not an exact solution yet */
+#define F_sigset "0x%06lx"
+typedef unsigned long T_sigset;
+
 /* default: socklen_t */
 #if !defined(HAVE_TYPEOF_STRUCT_CMSGHDR_CMSG_LEN) || !HAVE_TYPEOF_STRUCT_CMSGHDR_CMSG_LEN
 #  undef HAVE_TYPEOF_STRUCT_CMSGHDR_CMSG_LEN
@@ -651,6 +820,24 @@ typedef int sig_atomic_t;
 #  endif
 #endif
 
+/* basic type of struct timeval tv_usec */
+#ifndef F_tv_usec
+#  if TYPEOF_STRUCT_TIMEVAL_TV_USEC==1
+#     define F_tv_usec "%hu"
+#  elif TYPEOF_STRUCT_TIMEVAL_TV_USEC==3
+#     define F_tv_usec "%u"
+#  elif TYPEOF_STRUCT_TIMEVAL_TV_USEC==5
+#     define F_tv_usec "%lu"
+#  endif
+#endif
+
+/* OpenBSD (at least 7.2) does better with this special setting */
+#if __FreeBSD__ || __OpenBSD__
+#  define UNIX_TIGHTSOCKLEN false
+#else
+#  define UNIX_TIGHTSOCKLEN true
+#endif
+
 /* Cygwin 1.3.22 has the prototypes, but not the type... */
 #ifndef HAVE_TYPE_STAT64
 #  undef HAVE_STAT64
@@ -675,4 +862,8 @@ typedef int sig_atomic_t;
 extern const char *hstrerror(int);
 #endif
 
+#if !HAVE_PROTOTYPE_LIB_strndup
+extern char *strndup (const char *s, size_t n);
+#endif
+
 #endif /* !defined(__compat_h_included) */
