@@ -22,10 +22,11 @@
 #include "xioopts.h"
 #include "xiolockfile.h"
 
+#include "xio-pipe.h"
+
 
 /* command line options */
-struct {
-   size_t bufsiz;
+struct socat_opts {
    bool verbose;
    bool verbhex;
    struct timeval pollintv;	/* with ignoreeof, reread after seconds */
@@ -37,8 +38,9 @@ struct {
    bool lefttoright;	/* first addr ro, second addr wo */
    bool righttoleft;	/* first addr wo, second addr ro */
    xiolock_t lock;	/* a lock file */
+   unsigned long log_sigs;	/* signals to be caught just for logging */
+   bool statistics; 	/* log statistics on exit */
 } socat_opts = {
-   8192,	/* bufsiz */
    false,	/* verbose */
    false,	/* verbhex */
    {1,0},	/* pollintv */
@@ -50,14 +52,18 @@ struct {
    false,	/* lefttoright */
    false,	/* righttoleft */
    { NULL, 0 },	/* lock */
+   1<<SIGHUP | 1<<SIGINT | 1<<SIGQUIT | 1<<SIGILL | 1<<SIGABRT | 1<<SIGBUS | 1<<SIGFPE | 1<<SIGSEGV | 1<<SIGTERM, 	/* log_sigs */
+   false	/* statistics */
 };
 
 void socat_usage(FILE *fd);
+void socat_opt_hint(FILE *fd, char a, char b);
 void socat_version(FILE *fd);
 int socat(const char *address1, const char *address2);
 int _socat(void);
-int cv_newline(unsigned char **buff, ssize_t *bytes, int lineterm1, int lineterm2);
+int cv_newline(unsigned char *buff, ssize_t *bytes, int lineterm1, int lineterm2);
 void socat_signal(int sig);
+void socat_signal_logstats(int sig);
 static int socat_sigchild(struct single *file);
 
 void lftocrlf(char **in, ssize_t *len, size_t bufsiz);
@@ -66,6 +72,7 @@ void crlftolf(char **in, ssize_t *len, size_t bufsiz);
 static int socat_lock(void);
 static void socat_unlock(void);
 static int socat_newchild(void);
+static void socat_print_stats(void);
 
 static const char socatversion[] =
 #include "./VERSION"
@@ -80,13 +87,14 @@ const char copyright_ssleay[] = "This product includes software written by Tim H
 
 bool havelock;
 
-
 int main(int argc, const char *argv[]) {
    const char **arg1, *a;
    char *mainwaitstring;
    char buff[10];
    double rto;
    int i, argc0, result;
+   bool isdash = false;
+   int msglevel = 0;
    struct utsname ubuf;
    int lockrc;
 
@@ -108,7 +116,8 @@ int main(int argc, const char *argv[]) {
    arg1 = argv+1;  --argc;
    while (arg1[0] && (arg1[0][0] == '-')) {
       switch (arg1[0][1]) {
-      case 'V': socat_version(stdout); Exit(0);
+      case 'V':  if (arg1[0][2])  { socat_usage(stderr); Exit(1); }
+	 socat_version(stdout); Exit(0);
 #if WITH_HELP
       case '?':
       case 'h':
@@ -116,9 +125,43 @@ int main(int argc, const char *argv[]) {
 	 xioopenhelp(stdout, (arg1[0][2]=='?'||arg1[0][2]=='h') ? (arg1[0][3]=='?'||arg1[0][3]=='h') ? 2 : 1 : 0);
 	 Exit(0);
 #endif /* WITH_HELP */
-      case 'd': diag_set('d', NULL); break;
+      case 'd':
+	 a = *arg1+2;
+	 switch (*a) {
+	 case 'd':
+	    break;
+	 case '-': case '0': case '1': case '2': case '3': case '4':
+	    {
+	       char *endptr;
+	       msglevel = strtol(a, &endptr, 0);
+	       if (endptr == a || *endptr) {
+		  Error2("Invalid (trailing) character(s) \"%c\" in \"%s\"option", *a, *arg1);
+	       }
+	       diag_set_int('d', 4-msglevel);
+	    }
+	    break;
+	 case '\0':
+	    ++msglevel;
+	    diag_set_int('d', 4-msglevel);
+	    break;
+	 default: socat_usage(stderr);
+	 }
+	 if (*a != 'd')  break;
+	 ++msglevel;
+	 while (*a)  {
+	    if (*a == 'd') {
+	       ++msglevel;
+	       diag_set_int('d', 4-msglevel);
+	    } else {
+	       socat_usage(stderr);
+	       Exit(1);
+	    }
+	    ++a;
+	 }
+	 break;
 #if WITH_FILAN
-      case 'D': socat_opts.debug = true; break;
+      case 'D':  if (arg1[0][2])  { socat_opt_hint(stderr, arg1[0][1], arg1[0][2]); Exit(1); }
+	 socat_opts.debug = true; break;
 #endif
       case 'l':
 	 switch (arg1[0][2]) {
@@ -156,8 +199,32 @@ int main(int argc, const char *argv[]) {
 	    break;
 	 }
 	 break;
-      case 'v': socat_opts.verbose = true; break;
-      case 'x': socat_opts.verbhex = true; break;
+      case 'v':  if (arg1[0][2])  { socat_opt_hint(stderr, arg1[0][1], arg1[0][2]); Exit(1); }
+	 socat_opts.verbose = true; break;
+      case 'x':  if (arg1[0][2])  { socat_opt_hint(stderr, arg1[0][1], arg1[0][2]); Exit(1); }
+	 socat_opts.verbhex = true; break;
+      case 'r': if (arg1[0][2]) {
+	    a = *arg1+2;
+	 } else {
+	    ++arg1, --argc;
+	    if ((a = *arg1) == NULL) {
+	       Error("option -r requires an argument; use option \"-h\" for help");
+	       break;
+	    }
+	 }
+	 xiosetopt('r', a);
+	 break;
+      case 'R': if (arg1[0][2]) {
+	    a = *arg1+2;
+	 } else {
+	    ++arg1, --argc;
+	    if ((a = *arg1) == NULL) {
+	       Error("option -R requires an argument; use option \"-h\" for help");
+	       break;
+	    }
+	 }
+	 xiosetopt('R', a);
+	 break;
       case 'b': if (arg1[0][2]) {
 	    a = *arg1+2;
 	 } else {
@@ -167,10 +234,22 @@ int main(int argc, const char *argv[]) {
 	       Exit(1);
 	    }
 	 }
-	 socat_opts.bufsiz = strtoul(a, (char **)&a, 0);
+	 xioparms.bufsiz = Strtoul(a, (char **)&a, 0, "-b");
 	 break;
-      case 's':
+      case 's':  if (arg1[0][2])  { socat_opt_hint(stderr, arg1[0][1], arg1[0][2]); Exit(1); }
 	 diag_set_int('e', E_FATAL); break;
+      case 'S': 	/* do not catch signals */
+	 if (arg1[0][2]) {
+	    a = *arg1+2;
+	 } else {
+	    ++arg1, --argc;
+	    if ((a = *arg1) == NULL) {
+	       Error("option -S requires an argument; use option \"-h\" for help");
+	       Exit(1);
+	    }
+	 }
+	 socat_opts.log_sigs = Strtoul(a, (char **)&a, 0, "-S");
+	 break;
       case 't': if (arg1[0][2]) {
 	    a = *arg1+2;
 	 } else {
@@ -180,10 +259,10 @@ int main(int argc, const char *argv[]) {
 	       Exit(1);
 	    }
 	 }
-	 rto = strtod(a, (char **)&a);
+	 rto = Strtod(a, (char **)&a, "-t");
 	 socat_opts.closwait.tv_sec = rto;
 	 socat_opts.closwait.tv_usec =
-	    (rto-socat_opts.closwait.tv_sec) * 1000000; 
+	    (rto-socat_opts.closwait.tv_sec) * 1000000;
 	 break;
       case 'T':  if (arg1[0][2]) {
 	    a = *arg1+2;
@@ -194,14 +273,17 @@ int main(int argc, const char *argv[]) {
 	       Exit(1);
 	    }
 	 }
-	 rto = strtod(a, (char **)&a);
+	 rto = Strtod(a, (char **)&a, "-T");
 	 socat_opts.total_timeout.tv_sec = rto;
 	 socat_opts.total_timeout.tv_usec =
-	    (rto-socat_opts.total_timeout.tv_sec) * 1000000; 
+	    (rto-socat_opts.total_timeout.tv_sec) * 1000000;
 	 break;
-      case 'u': socat_opts.lefttoright = true; break;
-      case 'U': socat_opts.righttoleft = true; break;
-      case 'g': xioopts_ignoregroups = true; break;
+      case 'u':  if (arg1[0][2])  { socat_opt_hint(stderr, arg1[0][1], arg1[0][2]); Exit(1); }
+	 socat_opts.lefttoright = true; break;
+      case 'U':  if (arg1[0][2])  { socat_opt_hint(stderr, arg1[0][1], arg1[0][2]); Exit(1); }
+	 socat_opts.righttoleft = true; break;
+      case 'g':  if (arg1[0][2])  { socat_opt_hint(stderr, arg1[0][1], arg1[0][2]); Exit(1); }
+	 xioopts_ignoregroups = true; break;
       case 'L': if (socat_opts.lock.lockfile)
 	     Error("only one -L and -W option allowed");
 	 if (arg1[0][2]) {
@@ -214,8 +296,9 @@ int main(int argc, const char *argv[]) {
 	    }
 	 }
 	 break;
-      case 'W': if (socat_opts.lock.lockfile)
+      case 'W': if (socat_opts.lock.lockfile) {
 	    Error("only one -L and -W option allowed");
+	 }
 	 if (arg1[0][2]) {
 	    socat_opts.lock.lockfile = *arg1+2;
 	 } else {
@@ -236,27 +319,38 @@ int main(int argc, const char *argv[]) {
 #if WITH_IP6
       case '6':
 #endif
-	 xioopts.default_ip = arg1[0][1];
-	 xioopts.preferred_ip = arg1[0][1];
+	 if (arg1[0][2])  { socat_opt_hint(stderr, arg1[0][1], arg1[0][2]); Exit(1); }
+	 xioparms.default_ip = arg1[0][1];
+	 xioparms.preferred_ip = arg1[0][1];
 	 break;
 #endif /* WITH_IP4 || WITH_IP6 */
+      case '-':
+	 if (!strcmp("experimental", &arg1[0][2])) {
+	    xioparms.experimental = true;
+	 } else if (!strcmp("statistics", &arg1[0][2])) {
+	    socat_opts.statistics = true;
+	 } else {
+	    Error1("unknown option \"%s\"; use option \"-h\" for help", arg1[0]);
+	 }
+	 break;
       case '\0':
       case ',':
-      case ':': break;	/* this "-" is a variation of STDIO */
+      case ':':
+	 isdash = true;
+	 break;	/* this "-" is a variation of STDIO */
       default:
-	 xioinqopt('p', buff, sizeof(buff));
+	 xioinqopt('p', buff, sizeof(buff)); 	/* fetch pipe separator char */
 	 if (arg1[0][1] == buff[0]) {
+	    isdash = true;
 	    break;
 	 }
 	 Error1("unknown option \"%s\"; use option \"-h\" for help", arg1[0]);
 	 Exit(1);
       }
-      /* the leading "-" might be a form of the first address */
-      xioinqopt('p', buff, sizeof(buff));
-      if (arg1[0][0] == '-' &&
-	  (arg1[0][1] == '\0' || arg1[0][1] == ':' ||
-	   arg1[0][1] == ',' || arg1[0][1] == buff[0]))
+      if (isdash) {
+	 /* the leading "-" is a form of the first address */
 	 break;
+      }
       ++arg1; --argc;
    }
    if (argc != 2) {
@@ -286,20 +380,33 @@ int main(int argc, const char *argv[]) {
 #endif /* WITH_MSGLEVEL <= E_DEBUG */
 
    {
+#if HAVE_SIGACTION
       struct sigaction act;
-      sigfillset(&act.sa_mask);
+#endif
+      int i, m;
+
+      sigfillset(&act.sa_mask); 	/* while in sighandler block all signals */
       act.sa_flags = 0;
       act.sa_handler = socat_signal;
       /* not sure which signals should be caught and print a message */
-      Sigaction(SIGHUP,  &act, NULL);
-      Sigaction(SIGINT,  &act, NULL);
-      Sigaction(SIGQUIT, &act, NULL);
-      Sigaction(SIGILL,  &act, NULL);
-      Sigaction(SIGABRT, &act, NULL);
-      Sigaction(SIGBUS,  &act, NULL);
-      Sigaction(SIGFPE,  &act, NULL);
-      Sigaction(SIGSEGV, &act, NULL);
-      Sigaction(SIGTERM, &act, NULL);
+      for (i = 0, m = 1; i < 8*sizeof(unsigned long); ++i, m <<= 1) {
+	 if (socat_opts.log_sigs & m) {
+#if HAVE_SIGACTION
+	    Sigaction(i,  &act, NULL);
+#else
+	    Signal(i, socat_signal);
+#endif
+	 }
+      }
+
+#if WITH_STATS
+#if HAVE_SIGACTION
+      act.sa_handler = socat_signal_logstats;
+      Sigaction(SIGUSR1, &act, NULL);
+#else
+      Signal(SIGUSR1, socat_signal_logstats);
+#endif
+#endif /* WITH_STATS */
    }
    Signal(SIGPIPE, SIG_IGN);
 
@@ -314,8 +421,16 @@ int main(int argc, const char *argv[]) {
    }
 
    Atexit(socat_unlock);
+#if WITH_STATS
+   if (socat_opts.statistics) {
+      Atexit(socat_print_stats);
+   }
+#endif /* WITH_STATS */
 
    result = socat(arg1[0], arg1[1]);
+   if (result == EXIT_SUCCESS && engine_result != EXIT_SUCCESS) {
+      result = engine_result; 	/* a signal handler reports failure */
+   }
    Notice1("exiting with status %d", result);
    Exit(result);
    return 0;	/* not reached, just for gcc -Wall */
@@ -326,28 +441,34 @@ void socat_usage(FILE *fd) {
    fputs(copyright_socat, fd); fputc('\n', fd);
    fputs("Usage:\n", fd);
    fputs("socat [options] <bi-address> <bi-address>\n", fd);
-   fputs("   options:\n", fd);
+   fputs("   options (general command line options):\n", fd);
    fputs("      -V     print version and feature information to stdout, and exit\n", fd);
 #if WITH_HELP
    fputs("      -h|-?  print a help text describing command line options and addresses\n", fd);
    fputs("      -hh    like -h, plus a list of all common address option names\n", fd);
    fputs("      -hhh   like -hh, plus a list of all available address option names\n", fd);
 #endif /* WITH_HELP */
-   fputs("      -d     increase verbosity (use up to 4 times; 2 are recommended)\n", fd);
+   fputs("      -d[ddd]        increase verbosity (use up to 4 times; 2 are recommended)\n", fd);
+   fputs("      -d0|1|2|3|4    set verbosity level (0: Errors; 4 all including Debug)\n", fd);
 #if WITH_FILAN
    fputs("      -D     analyze file descriptors before loop\n", fd);
 #endif
+   fputs("      --experimental enable experimental features\n", fd);
+   fputs("      --statistics   output transfer statistics on exit\n", fd);
    fputs("      -ly[facility]  log to syslog, using facility (default is daemon)\n", fd);
    fputs("      -lf<logfile>   log to file\n", fd);
    fputs("      -ls            log to stderr (default if no other log)\n", fd);
    fputs("      -lm[facility]  mixed log mode (stderr during initialization, then syslog)\n", fd);
-   fputs("      -lp<progname>  set the program name used for logging\n", fd);
+   fputs("      -lp<progname>  set the program name used for logging and vars\n", fd);
    fputs("      -lu            use microseconds for logging timestamps\n", fd);
    fputs("      -lh            add hostname to log messages\n", fd);
-   fputs("      -v     verbose data traffic, text\n", fd);
-   fputs("      -x     verbose data traffic, hexadecimal\n", fd);
+   fputs("      -v     verbose text dump of data traffic\n", fd);
+   fputs("      -x     verbose hexadecimal dump of data traffic\n", fd);
+   fputs("      -r <file>      raw dump of data flowing from left to right\n", fd);
+   fputs("      -R <file>      raw dump of data flowing from right to left\n", fd);
    fputs("      -b<size_t>     set data buffer size (8192)\n", fd);
    fputs("      -s     sloppy (continue on error)\n", fd);
+   fputs("      -S<sigmask>    log these signals, override default\n", fd);
    fputs("      -t<timeout>    wait seconds before closing second channel\n", fd);
    fputs("      -T<timeout>    total inactivity timeout in seconds\n", fd);
    fputs("      -u     unidirectional mode (left to right)\n", fd);
@@ -363,6 +484,11 @@ void socat_usage(FILE *fd) {
 #endif
 }
 
+void socat_opt_hint(FILE *fd, char a, char b) {
+   fprintf(fd, "Do not merge single character options, i.e. use \"-%c -%c\" instead of \"-%c%c\"\n",
+	   a, b, a, b);
+}
+
 
 void socat_version(FILE *fd) {
    struct utsname ubuf;
@@ -373,6 +499,16 @@ void socat_version(FILE *fd) {
    fprintf(fd, "   running on %s version %s, release %s, machine %s\n",
 	   ubuf.sysname, ubuf.version, ubuf.release, ubuf.machine);
    fputs("features:\n", fd);
+#ifdef WITH_HELP
+   fprintf(fd, "  #define WITH_HELP %d\n", WITH_HELP);
+#else
+   fputs("  #undef WITH_HELP\n", fd);
+#endif
+#ifdef WITH_STATS
+   fprintf(fd, "  #define WITH_STATS %d\n", WITH_STATS);
+#else
+   fputs("  #undef WITH_STATS\n", fd);
+#endif
 #ifdef WITH_STDIO
    fprintf(fd, "  #define WITH_STDIO %d\n", WITH_STDIO);
 #else
@@ -408,6 +544,11 @@ void socat_version(FILE *fd) {
 #else
    fputs("  #undef WITH_PIPE\n", fd);
 #endif
+#ifdef WITH_SOCKETPAIR
+   fprintf(fd, "  #define WITH_SOCKETPAIR %d\n", WITH_SOCKETPAIR);
+#else
+   fputs("  #undef WITH_SOCKETPAIR\n", fd);
+#endif
 #ifdef WITH_UNIX
    fprintf(fd, "  #define WITH_UNIX %d\n", WITH_UNIX);
 #else
@@ -458,11 +599,26 @@ void socat_version(FILE *fd) {
 #else
    fputs("  #undef WITH_SCTP\n", fd);
 #endif
+#ifdef WITH_DCCP
+   fprintf(fd, "  #define WITH_DCCP %d\n", WITH_DCCP);
+#else
+   fputs("  #undef WITH_DCCP\n", fd);
+#endif
+#ifdef WITH_UDPLITE
+   fprintf(fd, "  #define WITH_UDPLITE %d\n", WITH_UDPLITE);
+#else
+   fputs("  #undef WITH_UDPLITE\n", fd);
+#endif
 #ifdef WITH_LISTEN
    fprintf(fd, "  #define WITH_LISTEN %d\n", WITH_LISTEN);
 #else
    fputs("  #undef WITH_LISTEN\n", fd);
 #endif
+#ifdef WITH_POSIXMQ
+   fprintf(fd, "  #define WITH_POSIXMQ %d\n", WITH_POSIXMQ);
+#else
+   fputs("  #undef WITH_POSIXMQ\n", fd);
+#endif
 #ifdef WITH_SOCKS4
    fprintf(fd, "  #define WITH_SOCKS4 %d\n", WITH_SOCKS4);
 #else
@@ -473,6 +629,21 @@ void socat_version(FILE *fd) {
 #else
    fputs("  #undef WITH_SOCKS4A\n", fd);
 #endif
+#ifdef WITH_SOCKS5
+   fprintf(fd, "  #define WITH_SOCKS5 %d\n", WITH_SOCKS5);
+#else
+   fputs("  #undef WITH_SOCKS5\n", fd);
+#endif
+#ifdef WITH_VSOCK
+   fprintf(fd, "  #define WITH_VSOCK %d\n", WITH_VSOCK);
+#else
+   fputs("  #undef WITH_VSOCK\n", fd);
+#endif
+#ifdef WITH_NAMESPACES
+   fprintf(fd, "  #define WITH_NAMESPACES %d\n", WITH_NAMESPACES);
+#else
+   fputs("  #undef WITH_NAMESPACES\n", fd);
+#endif
 #ifdef WITH_PROXY
    fprintf(fd, "  #define WITH_PROXY %d\n", WITH_PROXY);
 #else
@@ -483,6 +654,11 @@ void socat_version(FILE *fd) {
 #else
    fputs("  #undef WITH_SYSTEM\n", fd);
 #endif
+#ifdef WITH_SHELL
+   fprintf(fd, "  #define WITH_SHELL %d\n", WITH_SHELL);
+#else
+   fputs("  #undef WITH_SHELL\n", fd);
+#endif
 #ifdef WITH_EXEC
    fprintf(fd, "  #define WITH_EXEC %d\n", WITH_EXEC);
 #else
@@ -539,12 +715,23 @@ void socat_version(FILE *fd) {
 #else
    fputs("  #undef WITH_MSGLEVEL\n", fd);
 #endif
+#ifdef WITH_DEFAULT_IPV
+#  if WITH_DEFAULT_IPV
+   fprintf(fd, "  #define WITH_DEFAULT_IPV %c\n", WITH_DEFAULT_IPV);
+#  else
+   fprintf(fd, "  #define WITH_DEFAULT_IPV '\\0'\n");
+#  endif
+#else
+   fputs("  #undef WITH_DEFAULT_IPV\n", fd);
+#endif
 }
 
 
 xiofile_t *sock1, *sock2;
 int closing = 0;	/* 0..no eof yet, 1..first eof just occurred,
 			   2..counting down closing timeout */
+int sniffleft = -1; 		/* -1 or an FD for teeing data arriving on xfd1 */
+int sniffright = -1; 	/* -1 or an FD for teeing data arriving on xfd2 */
 
 /* call this function when the common command line options are parsed, and the
    addresses are extracted (but not resolved). */
@@ -572,29 +759,21 @@ int socat(const char *address1, const char *address2) {
        (XIO_RDSTREAM(sock1)->howtoend == END_KILL ||
 	XIO_RDSTREAM(sock1)->howtoend == END_CLOSE_KILL ||
 	XIO_RDSTREAM(sock1)->howtoend == END_SHUTDOWN_KILL)) {
-      if (XIO_RDSTREAM(sock1)->para.exec.pid == diedunknown1) {
-	 /* child has alread died... but it might have put regular data into
-	    the communication channel, so continue */
-	 Info1("child "F_pid" has already died (diedunknown1)",
-	       XIO_RDSTREAM(sock1)->para.exec.pid);
-	 diedunknown1 = 0;
-	 XIO_RDSTREAM(sock1)->para.exec.pid = 0;
-	 /* return STAT_RETRYLATER; */
-      } else if (XIO_RDSTREAM(sock1)->para.exec.pid == diedunknown2) {
-	 Info1("child "F_pid" has already died (diedunknown2)",
-	       XIO_RDSTREAM(sock1)->para.exec.pid);
-	 diedunknown2 = 0;
-	 XIO_RDSTREAM(sock1)->para.exec.pid = 0;
-      } else if (XIO_RDSTREAM(sock1)->para.exec.pid == diedunknown3) {
-	 Info1("child "F_pid" has already died (diedunknown3)",
-	       XIO_RDSTREAM(sock1)->para.exec.pid);
-	 diedunknown3 = 0;
-	 XIO_RDSTREAM(sock1)->para.exec.pid = 0;
-      } else if (XIO_RDSTREAM(sock1)->para.exec.pid == diedunknown4) {
-	 Info1("child "F_pid" has already died (diedunknown4)",
-	       XIO_RDSTREAM(sock1)->para.exec.pid);
-	 diedunknown4 = 0;
-	 XIO_RDSTREAM(sock1)->para.exec.pid = 0;
+      int i;
+      for (i = 0; i < NUMUNKNOWN; ++i) {
+	 if (XIO_RDSTREAM(sock1)->para.exec.pid == diedunknown[i]) {
+	    /* child has alread died... but it might have put regular data into
+	       the communication channel, so continue */
+	    Info2("child "F_pid" has already died with status %d",
+		  XIO_RDSTREAM(sock1)->para.exec.pid, statunknown[i]);
+	    ++num_child; 	/* it was counted as anonymous child, undo */
+	    if (statunknown[i] != 0) {
+	       return 1;
+	    }
+	    diedunknown[i] = 0;
+	    XIO_RDSTREAM(sock1)->para.exec.pid = 0;
+	    /* return STAT_RETRYLATER; */
+	 }
       }
    }
 #endif
@@ -623,36 +802,26 @@ int socat(const char *address1, const char *address2) {
        (XIO_RDSTREAM(sock2)->howtoend == END_KILL ||
 	XIO_RDSTREAM(sock2)->howtoend == END_CLOSE_KILL ||
 	XIO_RDSTREAM(sock2)->howtoend == END_SHUTDOWN_KILL)) {
-      if (XIO_RDSTREAM(sock2)->para.exec.pid == diedunknown1) {
-	 /* child has alread died... but it might have put regular data into
-	    the communication channel, so continue */
-	 Info1("child "F_pid" has already died (diedunknown1)",
-	       XIO_RDSTREAM(sock2)->para.exec.pid);
-	 diedunknown1 = 0;
-	 XIO_RDSTREAM(sock2)->para.exec.pid = 0;
-	 /* return STAT_RETRYLATER; */
-      } else if (XIO_RDSTREAM(sock2)->para.exec.pid == diedunknown2) {
-	 Info1("child "F_pid" has already died (diedunknown2)",
-	       XIO_RDSTREAM(sock2)->para.exec.pid);
-	 diedunknown2 = 0;
-	 XIO_RDSTREAM(sock2)->para.exec.pid = 0;
-      } else if (XIO_RDSTREAM(sock2)->para.exec.pid == diedunknown3) {
-	 Info1("child "F_pid" has already died (diedunknown3)",
-	       XIO_RDSTREAM(sock2)->para.exec.pid);
-	 diedunknown3 = 0;
-	 XIO_RDSTREAM(sock2)->para.exec.pid = 0;
-      } else if (XIO_RDSTREAM(sock2)->para.exec.pid == diedunknown4) {
-	 Info1("child "F_pid" has already died (diedunknown4)",
-	       XIO_RDSTREAM(sock2)->para.exec.pid);
-	 diedunknown4 = 0;
-	 XIO_RDSTREAM(sock2)->para.exec.pid = 0;
+      int i;
+      for (i = 0; i < NUMUNKNOWN; ++i) {
+	 if (XIO_RDSTREAM(sock2)->para.exec.pid == diedunknown[i]) {
+	    /* child has alread died... but it might have put regular data into
+	       the communication channel, so continue */
+	    Info2("child "F_pid" has already died with status %d",
+		  XIO_RDSTREAM(sock2)->para.exec.pid, statunknown[i]);
+	    if (statunknown[i] != 0) {
+	       return 1;
+	    }
+	    diedunknown[i] = 0;
+	    XIO_RDSTREAM(sock2)->para.exec.pid = 0;
+	    /* return STAT_RETRYLATER; */
+	 }
       }
    }
 #endif
 
    Info("resolved and opened all sock addresses");
-   return 
-      _socat();	/* nsocks, sockets are visible outside function */
+   return _socat();	/* nsocks, sockets are visible outside function */
 }
 
 /* checks if this is a connection to a child process, and if so, sees if the
@@ -701,7 +870,7 @@ int childleftdata(xiofile_t *xfd) {
 }
 
 int xiotransfer(xiofile_t *inpipe, xiofile_t *outpipe,
-		unsigned char **buff, size_t bufsiz, bool righttoleft);
+		unsigned char *buff, size_t bufsiz, bool righttoleft);
 
 bool mayrd1;		/* sock1 has read data or eof, according to poll() */
 bool mayrd2;		/* sock2 has read data or eof, according to poll() */
@@ -712,6 +881,7 @@ bool maywr2;		/* sock2 can be written to, according to poll() */
    and their options are set/applied
    returns -1 on error or 0 on success */
 int _socat(void) {
+   char *transferwaitstring;
    struct pollfd fds[4],
        *fd1in  = &fds[0],
        *fd1out = &fds[1],
@@ -724,6 +894,28 @@ int _socat(void) {
    int wasaction = 1;	/* last poll was active, do NOT sleep before next */
    struct timeval total_timeout;	/* the actual total timeout timer */
 
+   {
+      /* Open sniff file(s) */
+      char name[PATH_MAX];
+      struct timeval tv = { 0 }; 	/* 'cache' to have same time in both */
+
+      if (xioinqopt('r', name, sizeof(name)) == 0) {
+	 if (sniffleft >= 0)  Close(sniffleft);
+	 sniffleft = xio_opensnifffile(name, &tv);
+	 if (sniffleft < 0) {
+	    Error2("option -r \"%s\": %s", name, strerror(errno));
+         }
+      }
+
+      if (xioinqopt('R', name, sizeof(name)) == 0) {
+	 if (sniffright >= 0)  Close(sniffright);
+	 sniffright = xio_opensnifffile(name, &tv);
+	 if (sniffright < 0) {
+	    Error2("option -R \"%s\": %s", name, strerror(errno));
+         }
+      }
+   }
+
 #if WITH_FILAN
    if (socat_opts.debug) {
       int fdi, fdo;
@@ -754,16 +946,37 @@ int _socat(void) {
 #endif /* WITH_FILAN */
 
    /* when converting nl to crnl, size might double */
-   buff = Malloc(2*socat_opts.bufsiz+1);
+   if (xioparms.bufsiz > (SIZE_MAX-1)/2) {
+      Error2("buffer size option (-b) to big - "F_Zu" (max is "F_Zu")", xioparms.bufsiz, (SIZE_MAX-1)/2);
+      xioparms.bufsiz = (SIZE_MAX-1)/2;
+   }
+
+#if HAVE_PROTOTYPE_LIB_posix_memalign
+   /* Operations on files with flag O_DIRECT might need buffer alignment.
+      Without this, eg.read() fails with "Invalid argument" */
+   {
+      int _errno;
+      if ((_errno = Posix_memalign((void **)&buff, getpagesize(), 2*xioparms.bufsiz+1)) != 0) {
+	 Error1("posix_memalign(): %s", strerror(_errno));
+	 return -1;
+      }
+   }
+#else /* !HAVE_PROTOTYPE_LIB_posix_memalign */
+   buff = Malloc(2*xioparms.bufsiz+1);
    if (buff == NULL)  return -1;
+#endif /* !HAVE_PROTOTYPE_LIB_posix_memalign */
 
    if (socat_opts.logopt == 'm' && xioinqopt('l', NULL, 0) == 'm') {
       Info("switching to syslog");
-      diag_set('y', xioopts.syslogfac);
+      diag_set('y', xioparms.syslogfac);
       xiosetopt('l', "\0");
    }
    total_timeout = socat_opts.total_timeout;
 
+   if (transferwaitstring = getenv("SOCAT_TRANSFER_WAIT")) {
+      Info1("before starting data transfer loop: sleeping %ds (env:SOCAT_TRANSFER_WAIT)", atoi(transferwaitstring));
+      sleep(atoi(transferwaitstring));
+   }
    Notice4("starting data transfer loop with FDs [%d,%d] and [%d,%d]",
 	   XIO_GETRDFD(sock1), XIO_GETWRFD(sock1),
 	   XIO_GETRDFD(sock2), XIO_GETWRFD(sock2));
@@ -932,7 +1145,7 @@ int _socat(void) {
 	 if (closing) {
 	    break;
 	 }
-	 /* one possibility to come here is ignoreeof on some fd, but no EOF 
+	 /* one possibility to come here is ignoreeof on some fd, but no EOF
 	    and no data on any descriptor - this is no indication for end! */
 	 continue;
       }
@@ -942,7 +1155,7 @@ int _socat(void) {
 	 if (fd1in->revents & POLLNVAL) {
 	    /* this is what we find on Mac OS X when poll()'ing on a device or
 	       named pipe. a read() might imm. return with 0 bytes, resulting
-	       in a loop? */ 
+	       in a loop? */
 	    Error1("poll(...[%d]: invalid request", fd1in->fd);
 		  free(buff);
 	    return -1;
@@ -977,7 +1190,7 @@ int _socat(void) {
 
       if (mayrd1 && maywr2) {
 	 mayrd1 = false;
-	 if ((bytes1 = xiotransfer(sock1, sock2, &buff, socat_opts.bufsiz, false))
+	 if ((bytes1 = xiotransfer(sock1, sock2, buff, xioparms.bufsiz, false))
 	     < 0) {
 	    if (errno != EAGAIN) {
 	       closing = MAX(closing, 1);
@@ -1009,7 +1222,7 @@ int _socat(void) {
 
       if (mayrd2 && maywr1) {
 	 mayrd2 = false;
-	 if ((bytes2 = xiotransfer(sock2, sock1, &buff, socat_opts.bufsiz, true))
+	 if ((bytes2 = xiotransfer(sock2, sock1, buff, xioparms.bufsiz, true))
 	     < 0) {
 	    if (errno != EAGAIN) {
 	       closing = MAX(closing, 1);
@@ -1028,7 +1241,7 @@ int _socat(void) {
 		XIO_RDSTREAM(sock2)->actbytes == 0) {
 	       /* avoid idle when all readbytes already there */
 	       mayrd2 = true;
-	    }          
+	    }
 	    /* escape char occurred? */
 	    if (XIO_RDSTREAM(sock2)->actescape) {
 	       bytes2 = 0;	/* indicate EOF */
@@ -1051,10 +1264,10 @@ int _socat(void) {
 		   XIO_RDSTREAM(sock1)->fd);	/*! */
 	    mayrd1 = true;
 	    polling = 1;	/* do not hook this eof fd to poll for pollintv*/
-	 } else {
+	 } else if (XIO_RDSTREAM(sock1)->eof <= 2) {
 	    Notice1("socket 1 (fd %d) is at EOF", XIO_GETRDFD(sock1));
 	    xioshutdown(sock2, SHUT_WR);
-	    XIO_RDSTREAM(sock1)->eof = 2;
+	    XIO_RDSTREAM(sock1)->eof = 3;
 	    XIO_RDSTREAM(sock1)->ignoreeof = false;
 	 }
       } else if (polling && XIO_RDSTREAM(sock1)->ignoreeof) {
@@ -1074,10 +1287,10 @@ int _socat(void) {
 		   XIO_RDSTREAM(sock2)->fd);
 	    mayrd2 = true;
 	    polling = 1;	/* do not hook this eof fd to poll for pollintv*/
-	 } else {
+	 } else if (XIO_RDSTREAM(sock2)->eof <= 2) {
 	    Notice1("socket 2 (fd %d) is at EOF", XIO_GETRDFD(sock2));
 	    xioshutdown(sock1, SHUT_WR);
-	    XIO_RDSTREAM(sock2)->eof = 2;
+	    XIO_RDSTREAM(sock2)->eof = 3;
 	    XIO_RDSTREAM(sock2)->ignoreeof = false;
 	 }
       } else if (polling && XIO_RDSTREAM(sock2)->ignoreeof) {
@@ -1095,7 +1308,7 @@ int _socat(void) {
    xioclose(sock1);
    xioclose(sock2);
 
-		  free(buff);
+   free(buff);
    return 0;
 }
 
@@ -1106,40 +1319,45 @@ int _socat(void) {
    returns 0 on success or -1 if an error occurred */
 int gettimestamp(char *timestamp) {
    size_t bytes;
-#if HAVE_GETTIMEOFDAY || 1
+#if HAVE_CLOCK_GETTIME
+   struct timespec now;
+#elif HAVE_PROTOTYPE_LIB_gettimeofday
    struct timeval now;
-   int result;
+#endif /* !HAVE_PROTOTYPE_LIB_gettimeofday */
    time_t nowt;
-#else /* !HAVE_GETTIMEOFDAY */
-   time_t now;
-#endif /* !HAVE_GETTIMEOFDAY */
+   int result;
 
-#if HAVE_GETTIMEOFDAY || 1
-   result = gettimeofday(&now, NULL);
+#if HAVE_CLOCK_GETTIME
+   result = clock_gettime(CLOCK_REALTIME, &now);
    if (result < 0) {
       return result;
-   } else {
-      nowt = now.tv_sec;
-#if HAVE_STRFTIME
-      bytes = strftime(timestamp, 20, "%Y/%m/%d %H:%M:%S", localtime(&nowt));
-      bytes += sprintf(timestamp+19, "."F_tv_usec" ", now.tv_usec);
-#else
-      strcpy(timestamp, ctime(&nowt));
-      bytes = strlen(timestamp);
-#endif
    }
-#else /* !HAVE_GETTIMEOFDAY */
-   now = time(NULL);  if (now == (time_t)-1) {
+   nowt = now.tv_sec;
+#elif HAVE_PROTOTYPE_LIB_gettimeofday
+   result = Gettimeofday(&now, NULL);
+   if (result < 0) {
+      return result;
+   }
+   nowt = now.tv_sec;
+#else
+   nowt = time(NULL);
+   if (nowt == (time_t)-1) {
       return -1;
-   } else {
-#if HAVE_STRFTIME
-      bytes = strftime(timestamp, 21, "%Y/%m/%d %H:%M:%S ", localtime(&now));
-#else
-      strcpy(timestamp, ctime(&now));
-      bytes = strlen(timestamp);
-#endif
    }
-#endif /* !HAVE_GETTIMEOFDAY */
+#endif
+#if HAVE_STRFTIME
+   bytes = strftime(timestamp, 20, "%Y/%m/%d %H:%M:%S", localtime(&nowt));
+#if HAVE_CLOCK_GETTIME
+   bytes += sprintf(timestamp+19, "."F_tv_nsec" ", now.tv_nsec/1000);
+#elif HAVE_PROTOTYPE_LIB_gettimeofday
+   bytes += sprintf(timestamp+19, "."F_tv_usec" ", now.tv_usec);
+#else
+   strncpy(&timestamp[bytes++], " ", 2);
+#endif
+#else
+   strcpy(timestamp, ctime(&nowt));
+   bytes = strlen(timestamp);
+#endif
    return 0;
 }
 
@@ -1173,7 +1391,7 @@ static int
 /* inpipe is suspected to have read data available; read at most bufsiz bytes
    and transfer them to outpipe. Perform required data conversions.
    buff must be a malloc()'ed storage and might be realloc()'ed in this
-   function if more space is required after conversions. 
+   function if more space is required after conversions.
    Returns the number of bytes written, or 0 on EOF or <0 if an
    error occurred or when data was read but none written due to conversions
    (with EAGAIN). EAGAIN also occurs when reading from a nonblocking FD where
@@ -1183,10 +1401,11 @@ static int
    */
 /* inpipe, outpipe must be single descriptors (not dual!) */
 int xiotransfer(xiofile_t *inpipe, xiofile_t *outpipe,
-		unsigned char **buff, size_t bufsiz, bool righttoleft) {
+		unsigned char *buff, size_t bufsiz, bool righttoleft) {
    ssize_t bytes, writt = 0;
+   ssize_t sniffed;
 
-	 bytes = xioread(inpipe, *buff, bufsiz);
+	 bytes = xioread(inpipe, buff, bufsiz);
 	 if (bytes < 0) {
 	    if (errno != EAGAIN)
 	       XIO_RDSTREAM(inpipe)->eof = 2;
@@ -1201,10 +1420,14 @@ int xiotransfer(xiofile_t *inpipe, xiofile_t *outpipe,
 	 }
 
 	 if (bytes > 0) {
+#if WITH_STATS
+	    ++XIO_RDSTREAM(inpipe)->blocks_read;
+	    XIO_RDSTREAM(inpipe)->bytes_read += bytes;
+#endif
 	    /* handle escape char */
 	    if (XIO_RDSTREAM(inpipe)->escape != -1) {
 	       /* check input data for escape char */
-	       unsigned char *ptr = *buff;
+	       unsigned char *ptr = buff;
 	       size_t ctr = 0;
 	       while (ctr < bytes) {
 		  if (*ptr == XIO_RDSTREAM(inpipe)->escape) {
@@ -1234,14 +1457,34 @@ int xiotransfer(xiofile_t *inpipe, xiofile_t *outpipe,
 	       errno = EAGAIN;  return -1;
 	    }
 
+	    if (!righttoleft && sniffleft >= 0) {
+	       if ((sniffed = Write(sniffleft, buff, bytes)) < bytes) {
+		  if (sniffed < 0)
+		     Warn3("-r: write(%d, buff, "F_Zu"): %s",
+			   sniffleft, bytes, strerror(errno));
+		  else if (sniffed < bytes)
+		     Warn3("-r: write(%d, buff, "F_Zu") -> "F_Zd,
+			   sniffleft, bytes, sniffed);
+	       }
+	    } else if (righttoleft && sniffright >= 0) {
+	       if ((sniffed = Write(sniffright, buff, bytes)) < bytes) {
+		  if (sniffed < 0)
+		     Warn3("-R: write(%d, buff, "F_Zu"): %s",
+			   sniffright, bytes, strerror(errno));
+		  else if (sniffed < bytes)
+		     Warn3("-R: write(%d, buff, "F_Zu") -> "F_Zd,
+			   sniffright, bytes, sniffed);
+	       }
+	    }
+
 	    if (socat_opts.verbose && socat_opts.verbhex) {
 	       /* Hack-o-rama */
 	       size_t i = 0;
 	       size_t j;
 	       size_t N = 16;
 	       const unsigned char *end, *s, *t;
-	       s = *buff;
-	       end = (*buff)+bytes;
+	       s = buff;
+	       end = buff+bytes;
 	       xioprintblockheader(stderr, bytes, righttoleft);
 	       while (s < end) {
 		  /*! prefix? */
@@ -1287,8 +1530,8 @@ int xiotransfer(xiofile_t *inpipe, xiofile_t *outpipe,
 	       size_t i = 0;
 	       xioprintblockheader(stderr, bytes, righttoleft);
 	       while (i < (size_t)bytes) {
-		  int c = (*buff)[i];
-		  if (i > 0 && (*buff)[i-1] == '\n')
+		  int c = buff[i];
+		  if (i > 0 && buff[i-1] == '\n')
 		     /*! prefix? */;
 		  switch (c) {
 		  case '\a' : fputs("\\a", stderr); break;
@@ -1312,12 +1555,12 @@ int xiotransfer(xiofile_t *inpipe, xiofile_t *outpipe,
 	       /* print prefix */
 	       xioprintblockheader(stderr, bytes, righttoleft);
 	       for (i = 0; i < bytes; ++i) {
-		  fprintf(stderr, " %02x", (*buff)[i]);
+		  fprintf(stderr, " %02x", buff[i]);
 	       }
 	       fputc('\n', stderr);
 	    }
 
-	    writt = xiowrite(outpipe, *buff, bytes);
+	    writt = xiowrite(outpipe, buff, bytes);
 	    if (writt < 0) {
 	       /* EAGAIN when nonblocking but a mandatory lock is on file.
 		  the problem with EAGAIN is that the read cannot be repeated,
@@ -1332,6 +1575,10 @@ int xiotransfer(xiofile_t *inpipe, xiofile_t *outpipe,
 	    } else {
 	       Info3("transferred "F_Zu" bytes from %d to %d",
 		     writt, XIO_GETRDFD(inpipe), XIO_GETWRFD(outpipe));
+#if WITH_STATS
+	       ++XIO_WRSTREAM(outpipe)->blocks_written;
+	       XIO_WRSTREAM(outpipe)->bytes_written += writt;
+#endif
 	    }
 	 }
    return writt;
@@ -1344,11 +1591,9 @@ int xiotransfer(xiofile_t *inpipe, xiofile_t *outpipe,
 /* converts the newline characters (or character sequences) from the one
    specified in lineterm1 to that of lineterm2. Possible values are
    LINETERM_CR, LINETERM_CRNL, LINETERM_RAW.
-   buff points to the malloc()'ed data, input and output. It may be subject to
-   realloc(). bytes specifies the number of bytes input and output */
-int cv_newline(unsigned char **buff, ssize_t *bufsiz,
+   bytes specifies the number of bytes input and output */
+int cv_newline(unsigned char *buff, ssize_t *bytes,
 	       int lineterm1, int lineterm2) {
-   ssize_t *bytes = bufsiz;
    /* must perform newline changes */
    if (lineterm1 <= LINETERM_CR && lineterm2 <= LINETERM_CR) {
       /* no change in data length */
@@ -1358,23 +1603,23 @@ int cv_newline(unsigned char **buff, ssize_t *bufsiz,
       } else {
 	 from = '\r'; to = '\n';
       }
-      z = *buff + *bytes;
-      p = *buff;
+      z = buff + *bytes;
+      p = buff;
       while (p < z) {
 	 if (*p == from)  *p = to;
 	 ++p;
       }
 
    } else if (lineterm1 == LINETERM_CRNL) {
-      /* buffer becomes shorter */
+      /* buffer might become shorter */
       unsigned char to,  *s, *t, *z;
       if (lineterm2 == LINETERM_RAW) {
 	 to = '\n';
       } else {
 	 to = '\r';
       }
-      z = *buff + *bytes;
-      s = t = *buff;
+      z = buff + *bytes;
+      s = t = buff;
       while (s < z) {
 	 if (*s == '\r') {
 	    ++s;
@@ -1386,20 +1631,24 @@ int cv_newline(unsigned char **buff, ssize_t *bufsiz,
 	    *t++ = *s++;
 	 }
       }
-      *bufsiz = t - *buff;
+      *bytes = t - buff;
    } else {
-      /* buffer becomes longer, must alloc another space */
-      unsigned char *buf2;
+      /* buffer becomes longer (up to double length), must alloc another space */
+      static unsigned char *buf2;	/*! not threadsafe */
       unsigned char from;  unsigned char *s, *t, *z;
+
       if (lineterm1 == LINETERM_RAW) {
 	 from = '\n';
       } else {
 	 from = '\r';
       }
-      if ((buf2 = Malloc(2*socat_opts.bufsiz/*sic!*/+1)) == NULL) {
-	 return -1;
+      if (buf2 == NULL) {
+	 if ((buf2 = Malloc(xioparms.bufsiz)) == NULL) {
+	    return -1;
+	 }
       }
-      s = *buff;  t = buf2;  z = *buff + *bytes;
+      memcpy(buf2, buff, *bytes);
+      s = buf2;  t = buff;  z = buf2 + *bytes;
       while (s < z) {
 	 if (*s == from) {
 	    *t++ = '\r'; *t++ = '\n';
@@ -1409,9 +1658,7 @@ int cv_newline(unsigned char **buff, ssize_t *bufsiz,
 	    *t++ = *s++;
 	 }
       }
-      free(*buff);
-      *buff = buf2;
-      *bufsiz = t - buf2;;
+      *bytes = t - buff;;
    }
    return 0;
 }
@@ -1422,11 +1669,7 @@ void socat_signal(int signum) {
    diag_in_handler = 1;
    Notice1("socat_signal(): handling signal %d", signum);
    switch (signum) {
-   case SIGILL:
-   case SIGABRT:
-   case SIGBUS:
-   case SIGFPE:
-   case SIGSEGV:
+   default:
       diag_immediate_exit = 1;
    case SIGQUIT:
    case SIGPIPE:
@@ -1436,13 +1679,12 @@ void socat_signal(int signum) {
       break;
    case SIGTERM:
       Warn1("exiting on signal %d", signum); break;
-   case SIGHUP:  
+   case SIGHUP:
    case SIGINT:
       Notice1("exiting on signal %d", signum); break;
    }
-   //Exit(128+signum);
    Notice1("socat_signal(): finishing signal %d", signum);
-   diag_exit(128+signum);	/*!!! internal cleanup + _exit() */
+   diag_exit(128+signum);	/* internal cleanup + _exit() */
    diag_in_handler = 0;
    errno = _errno;
 }
@@ -1518,3 +1760,66 @@ static int socat_newchild(void) {
    havelock = false;
    return 0;
 }
+
+
+#if WITH_STATS
+void socat_signal_logstats(int signum) {
+   diag_in_handler = 1;
+   Notice1("socat_signal_logstats(): handling signal %d", signum);
+   socat_print_stats();
+   Notice1("socat_signal_logstats(): finishing signal %d", signum);
+   diag_in_handler = 0;
+}
+#endif /* WITH_STATS */
+
+#if WITH_STATS
+static void socat_print_stats(void)
+{
+	const char ltorf0[] = "STATISTICS: left to right: %%%ullu packets(s), %%%ullu byte(s)";
+	const char rtolf0[] = "STATISTICS: right to left: %%%ullu packets(s), %%%ullu byte(s)";
+	char ltorf1[sizeof(ltorf0)];	/* final printf format with lengths of number */
+	char rtolf1[sizeof(rtolf0)];	/* final printf format with lengths of number */
+	unsigned int blocksd = 1, bytesd = 1;	/* number of digits in output */
+	struct single *sock1w, *sock2w;
+	int savelevel;
+
+	if (sock1 == NULL || sock2 == NULL) {
+		Warn("transfer engine not yet started, statistics not available");
+		return;
+	}
+	if ((sock1->tag & ~XIO_TAG_CLOSED) == XIO_TAG_DUAL) {
+		sock1w = sock1->dual.stream[1];
+	} else {
+		sock1w = &sock1->stream;
+	}
+	if ((sock2->tag & ~XIO_TAG_CLOSED) == XIO_TAG_DUAL) {
+		sock2w = sock2->dual.stream[1];
+	} else {
+		sock2w = &sock2->stream;
+	}
+	if (!socat_opts.righttoleft && !socat_opts.righttoleft) {
+		/* Both directions - format output */
+		unsigned long long int maxblocks =
+			Max(sock1w->blocks_written, sock2w->blocks_written);
+		unsigned long long int maxbytes =
+			Max(sock1w->bytes_written,  sock2w->bytes_written);
+		/* Calculate number of digits */
+		while (maxblocks >= 10) { ++blocksd; maxblocks /= 10; }
+		while (maxbytes  >= 10) { ++bytesd;  maxbytes  /= 10; }
+	}
+	snprintf(ltorf1, sizeof(ltorf1), ltorf0, blocksd, bytesd);
+	snprintf(rtolf1, sizeof(rtolf1), rtolf0, blocksd, bytesd);
+	/* Statistics are E_INFO level; make sure they are printed anyway */
+	savelevel = diag_get_int('d');
+	diag_set_int('d', E_INFO);
+	Warn("statistics are experimental");
+	if (!socat_opts.righttoleft) {
+		Info2(ltorf1, sock2w->blocks_written, sock2w->bytes_written);
+	}
+	if (!socat_opts.lefttoright) {
+		Info2(rtolf1, sock1w->blocks_written, sock1w->bytes_written);
+	}
+	diag_set_int('d', savelevel);
+	return;
+}
+#endif /* WITH_STATs */
