@@ -1,5 +1,5 @@
 /* source: sycls.c */
-/* Copyright Gerhard Rieger */
+/* Copyright Gerhard Rieger and contributors (see file CHANGES) */
 /* Published under the GNU General Public License V.2, see file COPYING */
 
 /* explicit system call and C library trace function, for those who miss strace
@@ -8,8 +8,6 @@
 #include "config.h"
 #include "xioconfig.h"	/* what features are enabled */
 
-#if WITH_SYCLS
-
 #include "sysincludes.h"
 
 #include "mytypes.h"
@@ -23,6 +21,18 @@
 #include "sycls.h"
 
 
+#if WITH_SYCLS
+
+#if HAVE_PROTOTYPE_LIB_posix_memalign
+int Posix_memalign(void **memptr, size_t alignment, size_t size) {
+   int result;
+   Debug3("posix_memalign(%p, "F_Zu", "F_Zu")", memptr, alignment, size);
+   result = posix_memalign(memptr, alignment, size);
+   Debug1("posix_memalign(...) -> %d", result);
+   return result;
+}
+#endif /* HAVE_PROTOTYPE_LIB_posix_memalign */
+
 mode_t Umask(mode_t mask) {
    mode_t result;
    int _errno;
@@ -34,16 +44,27 @@ mode_t Umask(mode_t mask) {
    return result;
 }
 
+#endif /* WITH_SYCLS */
+
+
 int Open(const char *pathname, int flags, mode_t mode) {
    int result, _errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug3("open(\"%s\", 0%o, 0%03o)", pathname, flags, mode);
+#endif /* WITH_SYCLS */
    result = open(pathname, flags, mode);
    _errno = errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Info4("open(\"%s\", 0%o, 0%03o) -> %d", pathname, flags, mode, result);
+#endif /* WITH_SYCLS */
    errno = _errno;
    return result;
 }
 
+#if WITH_SYCLS
+
 int Creat(const char *pathname, mode_t mode) {
    int result, _errno;
    Debug2("creat(\"%s\", 0%03o)", pathname, mode);
@@ -336,8 +357,8 @@ int Gettimeofday(struct timeval *tv, struct timezone *tz) {
    }
 #endif /* WITH_MSGLEVEL <= E_DEBUG */
    result = gettimeofday(tv, tz);
-   _errno = errno;
 #if WITH_MSGLEVEL <= E_DEBUG
+   _errno = errno;
    if (tz) {
       Debug5("gettimeofday({%ld,%ld}, {%d,%d}) -> %d",
              tv->tv_sec, tv->tv_usec, tz->tz_minuteswest, tz->tz_dsttime,
@@ -346,8 +367,8 @@ int Gettimeofday(struct timeval *tv, struct timezone *tz) {
       Debug3("gettimeofday({%ld,%ld},) -> %d",
 	     tv->tv_sec, tv->tv_usec, result);
    }
-#endif /* WITH_MSGLEVEL <= E_DEBUG */
    errno = _errno;
+#endif /* WITH_MSGLEVEL <= E_DEBUG */
    return result;
 }
 
@@ -513,13 +534,25 @@ int Pipe(int filedes[2]) {
    return result;
 }
 
+#endif /* WITH_SYCLS */
+
 ssize_t Read(int fd, void *buf, size_t count) {
    ssize_t result;
    int _errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug3("read(%d, %p, "F_Zu")", fd, buf, count);
+#endif /* WITH_SYCLS */
    result = read(fd, buf, count);
    _errno = errno;
-   Debug1("read -> "F_Zd, result);
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
+   if (result < 0) {
+      Debug2("read -> "F_Zd" (errno=%d)", result, _errno);
+   } else {
+      Debug1("read -> "F_Zd, result);
+   }
+#endif /* WITH_SYCLS */
    errno = _errno;
    return result;
 }
@@ -527,45 +560,93 @@ ssize_t Read(int fd, void *buf, size_t count) {
 ssize_t Write(int fd, const void *buf, size_t count) {
    ssize_t result;
    int _errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug3("write(%d, %p, "F_Zu")", fd, buf, count);
+#endif /* WITH_SYCLS */
    result = write(fd, buf, count);
    _errno = errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug1("write -> "F_Zd, result);
+#endif /* WITH_SYCLS */
    errno = _errno;
    return result;
 }
 
+/* fcntl() without value */
 int Fcntl(int fd, int cmd) {
    int result, _errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug2("fcntl(%d, %d)", fd, cmd);
+#endif /* WITH_SYCLS */
    result = fcntl(fd, cmd);
+   if (!diag_in_handler)
+      diag_flush();
+#if WITH_SYCLS
    _errno = errno;
-   Debug1("fcntl() -> %d", result);
+   Debug1("fcntl() -> 0x%x", result);
+   errno = _errno;
+#endif /* WITH_SYCLS */
+   return result;
+}
+
+/* fcntl() with int value */
+int Fcntl_i(int fd, int cmd, int arg) {
+   int result, _errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
+   Debug3("fcntl(%d, %d, 0x%x)", fd, cmd, arg);
+#endif /* WITH_SYCLS */
+   result = fcntl(fd, cmd, arg);
+   _errno = errno;
+   if (!diag_in_handler)
+      diag_flush();
+#if WITH_SYCLS
+   Debug1("fcntl() -> 0x%x", result);
+#endif /* WITH_SYCLS */
    errno = _errno;
    return result;
 }
 
+/* fcntl() with long value */
 int Fcntl_l(int fd, int cmd, long arg) {
    int result, _errno;
-   Debug3("fcntl(%d, %d, %ld)", fd, cmd, arg);
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
+   Debug3("fcntl(%d, %d, 0x%lx)", fd, cmd, arg);
+#endif /* WITH_SYCLS */
    result = fcntl(fd, cmd, arg);
    _errno = errno;
-   Debug1("fcntl() -> %d", result);
+   if (!diag_in_handler)
+      diag_flush();
+#if WITH_SYCLS
+   Debug1("fcntl() -> 0x%x", result);
+#endif /* WITH_SYCLS */
    errno = _errno;
    return result;
 }
 
 int Fcntl_lock(int fd, int cmd, struct flock *l) {
    int result, _errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug7("fcntl(%d, %d, {type=%hd,whence=%hd,start="F_off",len="F_off",pid="F_pid"})",
 	  fd, cmd, l->l_type, l->l_whence, l->l_start, l->l_len, l->l_pid);
+#endif /* WITH_SYCLS */
    result = fcntl(fd, cmd, l);
    _errno = errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug1("fcntl() -> %d", result);
+#endif /* WITH_SYCLS */
    errno = _errno;
    return result;
 }
 
+#if WITH_SYCLS
+
 int Ftruncate(int fd, off_t length) {
    int retval, _errno;
    Debug2("ftruncate(%d, "F_off")", fd, length);
@@ -588,13 +669,21 @@ int Ftruncate64(int fd, off64_t length) {
 }
 #endif /* HAVE_FTRUNCATE64 */
 
+#endif /* WITH_SYCLS */
+
 #if HAVE_FLOCK
 int Flock(int fd, int operation) {
    int retval, _errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug2("flock(%d, %d)", fd, operation);
+#endif /* WITH_SYCLS */
    retval = flock(fd, operation);
    _errno = errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug1("flock() -> %d", retval);
+#endif /* WITH_SYCLS */
    errno = _errno;
    return retval;
 }
@@ -602,28 +691,45 @@ int Flock(int fd, int operation) {
 
 int Ioctl(int d, int request, void *argp) {
    int retval, _errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    if (argp > (void *)0x10000) {	/* fuzzy...*/
       Debug4("ioctl(%d, 0x%x, %p{%lu})", d, request, argp, *(unsigned long *)argp);
    } else {
       Debug3("ioctl(%d, 0x%x, 0x%p)", d, request, argp);
    }
+#endif /* WITH_SYCLS */
    retval = ioctl(d, request, argp);
    _errno = errno;
-   Debug1("ioctl() -> %d", retval);
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
+   if (retval < 0)
+      Debug2("ioctl() -> %d, errno=%d", retval, _errno);
+   else
+      Debug1("ioctl() -> %d", retval);
+#endif /* WITH_SYCLS */
    errno = _errno;
    return retval;
 }
 
 int Ioctl_int(int d, int request, int arg) {
    int retval, _errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug3("ioctl(%d, 0x%x, %d)", d, request, arg);
+#endif /* WITH_SYCLS */
    retval = ioctl(d, request, arg);
    _errno = errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug1("ioctl() -> %d", retval);
+#endif /* WITH_SYCLS */
    errno = _errno;
    return retval;
 }
 
+#if WITH_SYCLS
+
 int Close(int fd) {
    int retval, _errno;
    Info1("close(%d)", fd);
@@ -704,10 +810,14 @@ int Chmod(const char *path, mode_t mode) {
    return retval;
 }
 
+#endif /* WITH_SYCLS */
+
 #if HAVE_POLL
 /* we only show the first struct pollfd; hope this is enough for most cases. */
 int Poll(struct pollfd *ufds, unsigned int nfds, int timeout) {
-   int result;
+   int _errno, result;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    if (nfds == 4) {
       Debug10("poll({%d,0x%02hx,}{%d,0x%02hx,}{%d,0x%02hx,}{%d,0x%02hx,}, %u, %d)",
 	      ufds[0].fd, ufds[0].events, ufds[1].fd, ufds[1].events,
@@ -716,13 +826,19 @@ int Poll(struct pollfd *ufds, unsigned int nfds, int timeout) {
    } else {
       Debug4("poll({%d,0x%02hx,}, , %u, %d)", ufds[0].fd, ufds[0].events, nfds, timeout);
    }
+#endif /* WITH_SYCLS */
    result = poll(ufds, nfds, timeout);
+   _errno = errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    if (nfds == 4) {
       Debug5("poll(, {,,0x%02hx}{,,0x%02hx}{,,0x%02hx}{,,0x%02hx}) -> %d",
 	     ufds[0].revents, ufds[1].revents, ufds[2].revents, ufds[3].revents, result);
    } else {
       Debug2("poll(, {,,0x%02hx}) -> %d", ufds[0].revents, result);
    }
+#endif /* WITH_SYCLS */
+   errno = _errno;
    return result;
 }
 #endif /* HAVE_POLL */
@@ -732,37 +848,91 @@ int Poll(struct pollfd *ufds, unsigned int nfds, int timeout) {
 int Select(int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
 	   struct timeval *timeout) {
    int result, _errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
 #if HAVE_FDS_BITS
    Debug7("select(%d, &0x%lx, &0x%lx, &0x%lx, %s%lu."F_tv_usec")",
-	  n, readfds->fds_bits[0], writefds->fds_bits[0],
-	  exceptfds->fds_bits[0],
+	  n, readfds?readfds->fds_bits[0]:0, writefds?writefds->fds_bits[0]:0,
+	  exceptfds?exceptfds->fds_bits[0]:0,
 	  timeout?"&":"NULL/", timeout?timeout->tv_sec:0,
 	  timeout?timeout->tv_usec:0);
 #else
    Debug7("select(%d, &0x%lx, &0x%lx, &0x%lx, %s%lu.%06u)",
-	  n, readfds->__fds_bits[0], writefds->__fds_bits[0],
-	  exceptfds->__fds_bits[0],
+	  n, readfds?readfds->__fds_bits[0]:0, writefds?writefds->__fds_bits[0]:0,
+	  exceptfds?exceptfds->__fds_bits[0]:0,
 	  timeout?"&":"NULL/", timeout?timeout->tv_sec:0,
 	  timeout?timeout->tv_usec:0);
 #endif
+#endif /* WITH_SYCLS */
    result = select(n, readfds, writefds, exceptfds, timeout);
    _errno = errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
 #if HAVE_FDS_BITS
    Debug7("select -> (, 0x%lx, 0x%lx, 0x%lx, %s%lu."F_tv_usec"), %d",
-	  readfds->fds_bits[0], writefds->fds_bits[0], exceptfds->fds_bits[0],
+	  readfds?readfds->fds_bits[0]:0, writefds?writefds->fds_bits[0]:0,
+	  exceptfds?exceptfds->fds_bits[0]:0,
 	  timeout?"&":"NULL/", timeout?timeout->tv_sec:0,
 	  timeout?timeout->tv_usec:0, result);
 #else
    Debug7("select -> (, 0x%lx, 0x%lx, 0x%lx, %s%lu.%06u), %d",
-	  readfds->__fds_bits[0], writefds->__fds_bits[0],
-	  exceptfds->__fds_bits[0],
+	  readfds?readfds->__fds_bits[0]:0, writefds?writefds->__fds_bits[0]:0,
+	  exceptfds?exceptfds->__fds_bits[0]:0,
 	  timeout?"&":"NULL/", timeout?timeout->tv_sec:0,
 	  timeout?timeout->tv_usec:0, result);
 #endif
+#endif /* WITH_SYCLS */
    errno = _errno;
+
    return result;
 }
 
+#if HAVE_PSELECT
+/* we only show the first word of the fd_set's; hope this is enough for most
+   cases. */
+int Pselect(int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
+	    const struct timespec *timeout, const sigset_t *sigmask) {
+   int result, _errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
+#if HAVE_FDS_BITS
+   Debug8("pselect(%d, &0x%lx, &0x%lx, &0x%lx, %s%lu."F_tv_nsec", "F_sigset")",
+	  n, readfds?readfds->fds_bits[0]:0, writefds?writefds->fds_bits[0]:0,
+	  exceptfds?exceptfds->fds_bits[0]:0,
+	  timeout?"&":"NULL/", timeout?timeout->tv_sec:0,
+	  timeout?timeout->tv_nsec:0, *(T_sigset *)sigmask);
+#else
+   Debug8("pselect(%d, &0x%lx, &0x%lx, &0x%lx, %s%lu.%06u)",
+	  n, readfds?readfds->__fds_bits[0]:0, writefds?writefds->__fds_bits[0]:0,
+	  exceptfds?exceptfds->__fds_bits[0]:0,
+	  timeout?"&":"NULL/", timeout?timeout->tv_sec:0,
+	  timeout?timeout->tv_nsec:0);
+#endif
+#endif /* WITH_SYCLS */
+   result = pselect(n, readfds, writefds, exceptfds, timeout, sigmask);
+   _errno = errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
+#if HAVE_FDS_BITS
+   Debug5("pselect -> (, 0x%lx, 0x%lx, 0x%lx), "F_sigset", %d",
+	  readfds?readfds->fds_bits[0]:0, writefds?writefds->fds_bits[0]:0,
+	  exceptfds?exceptfds->fds_bits[0]:0, *(T_sigset *)sigmask,
+	  result);
+#else
+   Debug6("pselect -> (, 0x%lx, 0x%lx, 0x%lx), %d",
+	  readfds?readfds->__fds_bits[0]:0, writefds?writefds->__fds_bits[0]:0,
+	  exceptfds?exceptfds->__fds_bits[0]:0,
+	  result);
+#endif
+#endif /* WITH_SYCLS */
+   errno = _errno;
+
+   return result;
+}
+#endif /* HAVE_PSELECT */
+
+#if WITH_SYCLS
+
 pid_t Fork(void) {
    pid_t pid;
    int _errno;
@@ -774,17 +944,27 @@ pid_t Fork(void) {
    return pid;
 }
 
+#endif /* WITH_SYCLS */
+
 pid_t Waitpid(pid_t pid, int *status, int options) {
    int _errno;
    pid_t retval;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug3("waitpid("F_pid", %p, %d)", pid, status, options);
+#endif /* WITH_SYCLS */
    retval = waitpid(pid, status, options);
    _errno = errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug2("waitpid(, {%d}, ) -> "F_pid, *status, retval);
+#endif /* WITH_SYCLS */
    errno = _errno;
    return retval;
 }
 
+#if WITH_SYCLS
+
 sighandler_t Signal(int signum, sighandler_t handler) {
    int _errno;
    sighandler_t retval;
@@ -809,9 +989,15 @@ int Sigaction(int signum, const struct sigaction *act,
 
 int Sigprocmask(int how, const sigset_t *set, sigset_t *oset) {
    int retval;
-   Debug3("sigprocmask(%d, %p, %p)", how, set, oset);
+   if (set)
+      Debug3("sigprocmask(%d, "F_sigset", %p)", how, *(T_sigset *)set, oset);
+   else
+      Debug2("sigprocmask(%d, NULL, %p)", how, oset);
    retval = sigprocmask(how, set, oset);
-   Debug1("sigprocmask() -> %d", retval);
+   if (oset)
+      Debug2("sigprocmask() -> {,, "F_sigset"} %d", *(T_sigset *)oset, retval);
+   else
+      Debug1("sigprocmask() -> %d", retval);
    return retval;
 }
 
@@ -865,16 +1051,26 @@ int Execvp(const char *file, char *const argv[]) {
    return result;
 }
 
+#endif /* WITH_SYCLS */
+
 int System(const char *string) {
    int result, _errno;
+#if WITH_SYCLS
    Debug1("system(\"%s\")", string);
+#endif /* WITH_SYCLS */
+   diag_immediate_exit = 1;
    result = system(string);
+   diag_immediate_exit = 0;
+#if WITH_SYCLS
    _errno = errno;
    Debug1("system() -> %d", result);
    errno = _errno;
+#endif /* WITH_SYCLS */
    return result;
 }
 
+#if WITH_SYCLS
+
 int Socketpair(int d, int type, int protocol, int sv[2]) {
    int result, _errno;
    Debug4("socketpair(%d, %d, %d, %p)", d, type, protocol, sv);
@@ -912,24 +1108,28 @@ int Bind(int sockfd, struct sockaddr *my_addr, socklen_t addrlen) {
 }
 #endif /* _WITH_SOCKET */
 
+#endif /* WITH_SYCLS */
+
 #if _WITH_SOCKET
 int Connect(int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen) {
    int result, _errno;
    char infobuff[256];
 
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    /*sockaddr_info(serv_addr, infobuff, sizeof(infobuff));
    Debug3("connect(%d, %s, "F_Zd")", sockfd, infobuff, addrlen);*/
 #if 0
    Debug18("connect(%d,{0x%02x%02x%02x%02x %02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x}, "F_Zd")",
 	   sockfd,
 	   ((unsigned char *)serv_addr)[0],  ((unsigned char *)serv_addr)[1],
-	   ((unsigned char *)serv_addr)[2],  ((unsigned char *)serv_addr)[3], 
-	   ((unsigned char *)serv_addr)[4],  ((unsigned char *)serv_addr)[5], 
-	   ((unsigned char *)serv_addr)[6],  ((unsigned char *)serv_addr)[7], 
-	   ((unsigned char *)serv_addr)[8],  ((unsigned char *)serv_addr)[9], 
-	   ((unsigned char *)serv_addr)[10], ((unsigned char *)serv_addr)[11], 
-	   ((unsigned char *)serv_addr)[12], ((unsigned char *)serv_addr)[13], 
-	   ((unsigned char *)serv_addr)[14], ((unsigned char *)serv_addr)[15], 
+	   ((unsigned char *)serv_addr)[2],  ((unsigned char *)serv_addr)[3],
+	   ((unsigned char *)serv_addr)[4],  ((unsigned char *)serv_addr)[5],
+	   ((unsigned char *)serv_addr)[6],  ((unsigned char *)serv_addr)[7],
+	   ((unsigned char *)serv_addr)[8],  ((unsigned char *)serv_addr)[9],
+	   ((unsigned char *)serv_addr)[10], ((unsigned char *)serv_addr)[11],
+	   ((unsigned char *)serv_addr)[12], ((unsigned char *)serv_addr)[13],
+	   ((unsigned char *)serv_addr)[14], ((unsigned char *)serv_addr)[15],
 	   addrlen);
 #else
    Debug4("connect(%d, {%d,%s}, "F_socklen")",
@@ -937,14 +1137,20 @@ int Connect(int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen) {
 	  sockaddr_info(serv_addr, addrlen, infobuff, sizeof(infobuff)),
 	  addrlen);
 #endif
+#endif /* WITH_SYCLS */
    result = connect(sockfd, serv_addr, addrlen);
    _errno = errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug1("connect() -> %d", result);
+#endif /* WITH_SYCLS */
    errno = _errno;
    return result;
 }
 #endif /* _WITH_SOCKET */
 
+#if WITH_SYCLS
+
 #if _WITH_SOCKET
 int Listen(int s, int backlog) {
    int result, _errno;
@@ -957,17 +1163,29 @@ int Listen(int s, int backlog) {
 }
 #endif /* _WITH_SOCKET */
 
+#endif /* WITH_SYCLS */
+
 #if _WITH_SOCKET
 /* don't forget to handle EINTR when using Accept() ! */
 int Accept(int s, struct sockaddr *addr, socklen_t *addrlen) {
+      char infobuff[256];
    int result, _errno;
-
-   Debug3("accept(%d, %p, %p)", s, addr, addrlen);
+   fd_set accept_s;
+   if (!diag_in_handler) diag_flush();
+   FD_ZERO(&accept_s);
+   FD_SET(s, &accept_s);
+   if (diag_select(s+1, &accept_s, NULL, NULL, NULL) < 0) {
+      return -1;
+   }
+#if WITH_SYCLS
+   sockaddr_info(addr, *addrlen, infobuff, sizeof(infobuff));
+   Debug3("accept(%d, %p, %p)", s, infobuff, addrlen);
+#endif /* WITH_SYCLS */
    result = accept(s, addr, addrlen);
    _errno = errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    if (result >= 0) {
-      char infobuff[256];
-      sockaddr_info(addr, *addrlen, infobuff, sizeof(infobuff));
       Info5("accept(%d, {%d, %s}, "F_socklen") -> %d", s,
 	    addr->sa_family,
 	    sockaddr_info(addr, *addrlen, infobuff, sizeof(infobuff)),
@@ -975,11 +1193,14 @@ int Accept(int s, struct sockaddr *addr, socklen_t *addrlen) {
    } else {
       Debug1("accept(,,) -> %d", result);
    }
+#endif /* WITH_SYCLS */
    errno = _errno;
    return result;
 }
 #endif /* _WITH_SOCKET */
 
+#if WITH_SYCLS
+
 #if _WITH_SOCKET
 int Getsockname(int s, struct sockaddr *name, socklen_t *namelen) {
    int result, _errno;
@@ -1048,13 +1269,21 @@ int Setsockopt(int s, int level, int optname, const void *optval, int optlen) {
 }
 #endif /* _WITH_SOCKET */
 
+#endif /* WITH_SYCLS */
+
 #if _WITH_SOCKET
 int Recv(int s, void *buf, size_t len, int flags) {
    int retval, _errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug4("recv(%d, %p, "F_Zu", %d)", s, buf, len, flags);
+#endif /* WITH_SYCLS */
    retval = recv(s, buf, len, flags);
    _errno = errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug1("recv() -> %d", retval);
+#endif /* WITH_SYCLS */
    errno = _errno;
    return retval;
 }
@@ -1065,10 +1294,15 @@ int Recvfrom(int s, void *buf, size_t len, int flags, struct sockaddr *from,
 	     socklen_t *fromlen) {
    int retval, _errno;
    char infobuff[256];
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug6("recvfrom(%d, %p, "F_Zu", %d, %p, "F_socklen")",
 	  s, buf, len, flags, from, *fromlen);
+#endif /* WITH_SYCLS */
    retval = recvfrom(s, buf, len, flags, from, fromlen);
    _errno = errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    if (from) {
       Debug4("recvfrom(,,,, {%d,%s}, "F_socklen") -> %d",
 	     from->sa_family,
@@ -1077,6 +1311,7 @@ int Recvfrom(int s, void *buf, size_t len, int flags, struct sockaddr *from,
    } else {
       Debug1("recvfrom(,,,, NULL, NULL) -> %d", retval);
    }
+#endif /* WITH_SYCLS */
    errno = _errno;
    return retval;
 }
@@ -1085,6 +1320,8 @@ int Recvfrom(int s, void *buf, size_t len, int flags, struct sockaddr *from,
 #if _WITH_SOCKET
 int Recvmsg(int s, struct msghdr *msgh, int flags) {
    int retval, _errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    char infobuff[256];
 #if defined(HAVE_STRUCT_MSGHDR_MSGCONTROL) && defined(HAVE_STRUCT_MSGHDR_MSGCONTROLLEN) && defined(HAVE_STRUCT_MSGHDR_MSGFLAGS)
    Debug10("recvmsg(%d, %p{%p,%u,%p,"F_Zu",%p,"F_Zu",%d}, %d)", s, msgh,
@@ -1095,8 +1332,11 @@ int Recvmsg(int s, struct msghdr *msgh, int flags) {
 	  msgh->msg_name, msgh->msg_namelen,  msgh->msg_iov,  msgh->msg_iovlen,
 	  flags);
 #endif
+#endif /* WITH_SYCLS */
    retval = recvmsg(s, msgh, flags);
    _errno = errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
 #if defined(HAVE_STRUCT_MSGHDR_MSGCONTROLLEN)
    Debug5("recvmsg(, {%s,%u,,"F_Zu",,"F_Zu",}, ) -> %d",
 	  msgh->msg_name?sockaddr_info(msgh->msg_name, msgh->msg_namelen, infobuff, sizeof(infobuff)):"NULL",
@@ -1108,6 +1348,7 @@ int Recvmsg(int s, struct msghdr *msgh, int flags) {
 	  msgh->msg_namelen, msgh->msg_iovlen,
 	  retval);
 #endif
+#endif /* WITH_SYCLS */
    errno = _errno;
    return retval;
 }
@@ -1116,11 +1357,17 @@ int Recvmsg(int s, struct msghdr *msgh, int flags) {
 #if _WITH_SOCKET
 int Send(int s, const void *mesg, size_t len, int flags) {
    int retval, _errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug5("send(%d, %p[%08x...], "F_Zu", %d)",
 	  s, mesg, ntohl(*(unsigned long *)mesg), len, flags);
+#endif /* WITH_SYCLS */
    retval = send(s, mesg, len, flags);
    _errno = errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug1("send() -> %d", retval);
+#endif /* WITH_SYCLS */
    errno = _errno;
    return retval;
 }
@@ -1132,17 +1379,25 @@ int Sendto(int s, const void *mesg, size_t len, int flags,
    int retval, _errno;
    char infobuff[256];
 
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    sockaddr_info(to, tolen, infobuff, sizeof(infobuff));
    Debug7("sendto(%d, %p[%08x...], "F_Zu", %d, {%s}, %d)",
 	  s, mesg, htonl(*(unsigned long *)mesg), len, flags, infobuff, tolen);
+#endif /* WITH_SYCLS */
    retval = sendto(s, mesg, len, flags, to, tolen);
    _errno = errno;
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug1("sendto() -> %d", retval);
+#endif /* WITH_SYCLS */
    errno = _errno;
    return retval;
 }
 #endif /* _WITH_SOCKET */
 
+#if WITH_SYCLS
+
 #if _WITH_SOCKET
 int Shutdown(int fd, int how) {
    int retval, _errno;
@@ -1163,13 +1418,6 @@ unsigned int Sleep(unsigned int seconds) {
    return retval;
 }
 
-void Usleep(unsigned long usec) {
-   Debug1("usleep(%lu)", usec);
-   usleep(usec);
-   Debug("usleep() ->");
-   return;
-}
-
 #if HAVE_NANOSLEEP
 unsigned int Nanosleep(const struct timespec *req, struct timespec *rem) {
    int retval, _errno;
@@ -1197,7 +1445,7 @@ int Pause(void) {
    return retval;
 }
 
-#if WITH_IP4 || WITH_IP6
+#if ( _WITH_IP4 || _WITH_IP6 ) && HAVE_GETHOSTBYNAME
 struct hostent *Gethostbyname(const char *name) {
    struct hostent *hent;
    Debug1("gethostbyname(\"%s\")", name);
@@ -1213,13 +1461,13 @@ struct hostent *Gethostbyname(const char *name) {
    }
    return hent;
 }
-#endif /* WITH_IP4 || WITH_IP6 */
+#endif /* ( _WITH_IP4 || _WITH_IP6 ) && HAVE_GETHOSTBYNAME */
 
 #if (_WITH_IP4 || _WITH_IP6) && HAVE_GETADDRINFO
 int Getaddrinfo(const char *node, const char *service,
 		const struct addrinfo *hints, struct addrinfo **res) {
    int result;
-   Debug15("getaddrinfo(%s%s%s, %s%s%s, {%d,%d,%d,%d,"F_socklen",%p,%p,%p}, %p)",
+   Debug15("getaddrinfo(%s%s%s, %s%s%s, {0x%02x,%d,%d,%d,"F_socklen",%p,%p,%p}, %p)",
 	   node?"\"":"", node?node:"NULL", node?"\"":"",
 	   service?"\"":"", service?service:"NULL", service?"\"":"",
 	   hints->ai_flags, hints->ai_family, hints->ai_socktype,
@@ -1305,10 +1553,18 @@ int Tcgetattr(int fd, struct termios *termios_p) {
       cp += sprintf(cp, "%02x,", termios_p->c_cc[i]);
    }
    sprintf(cp, "%02x", termios_p->c_cc[i]);
+#if HAVE_STRUCT_TERMIOS_C_ISPEED && HAVE_STRUCT_TERMIOS_C_OSPEED
+   Debug8("tcgetattr(, {%08x,%08x,%08x,%08x, "F_speed","F_speed", %s}) -> %d",
+	  termios_p->c_iflag, termios_p->c_oflag,
+	  termios_p->c_cflag, termios_p->c_lflag,
+	  termios_p->c_ispeed, termios_p->c_ospeed,
+	  chars, result);
+#else
    Debug6("tcgetattr(, {%08x,%08x,%08x,%08x,%s}) -> %d",
 	  termios_p->c_iflag, termios_p->c_oflag,
-	  termios_p->c_cflag, termios_p->c_lflag, 
+	  termios_p->c_cflag, termios_p->c_lflag,
 	  chars, result);
+#endif
    errno = _errno;
    return result;
 }
@@ -1323,9 +1579,18 @@ int Tcsetattr(int fd, int optional_actions, struct termios *termios_p) {
       cp += sprintf(cp, "%02x,", termios_p->c_cc[i]);
    }
    sprintf(cp, "%02x", termios_p->c_cc[i]);
+#if HAVE_STRUCT_TERMIOS_C_ISPEED && HAVE_STRUCT_TERMIOS_C_OSPEED
+   Debug9("tcsetattr(%d, %d, {%08x,%08x,%08x,%08x, "F_speed","F_speed", %s})",
+	  fd, optional_actions,
+	  termios_p->c_iflag, termios_p->c_oflag,
+	  termios_p->c_cflag, termios_p->c_lflag,
+	  termios_p->c_ispeed, termios_p->c_ospeed,
+	  chars);
+#else
    Debug7("tcsetattr(%d, %d, {%08x,%08x,%08x,%08x,%s})", fd, optional_actions,
 	  termios_p->c_iflag, termios_p->c_oflag,
 	  termios_p->c_cflag, termios_p->c_lflag, chars);
+#endif
    result = tcsetattr(fd, optional_actions, termios_p);
    _errno = errno;
    Debug1("tcsetattr() -> %d", result);
@@ -1449,12 +1714,18 @@ int Atexit(void (*func)(void)) {
    return result;
 }
 
+#endif /* WITH_SYCLS */
 
 void Exit(int status) {
+   if (!diag_in_handler) diag_flush();
+#if WITH_SYCLS
    Debug1("exit(%d)", status);
+#endif /* WITH_SYCLS */
    exit(status);
 }
 
+#if WITH_SYCLS
+
 void Abort(void) {
    Debug("abort()");
    abort();
@@ -1494,6 +1765,28 @@ void Unsetenv(const char *name) {
 }
 #endif
 
+
+#if WITH_NAMESPACES
+int Setns(
+	int fd,
+	int nstype)
+{
+	int _errno;
+	int result;
+	Debug2("setns(%d, 0x%x)", fd, nstype);
+	result = setns(fd, nstype);
+	if (result < 0) {
+		_errno = errno;
+		Debug2("setns() -> %d (errno=%d)", result, errno);
+		errno = _errno;
+		return result;
+	}
+	Debug1("setns() -> %d", result);
+	return result;
+}
+#endif /* WITH_NAMESPACES */
+
+
 #if WITH_READLINE
 
 char *Readline(const char *prompt) {
