@@ -16,6 +16,10 @@
 #include "utils.h"
 #include "sysutils.h"
 
+#if _WITH_INTERFACE
+const int one = 1;
+#endif
+
 /* Substitute for Write():
    Try to write all bytes before returning; this handles EINTR,
    EAGAIN/EWOULDBLOCK, and partial write situations. The drawback is that this
@@ -40,7 +44,7 @@ ssize_t writefull(int fd, const void *buff, size_t bytes) {
 	 default: return -1;
 	 }
       } else if (writt+chk < bytes) {
-	 Warn4("write(%d, %p, "F_Zu"): only wrote "F_Zu" bytes, trying to continue ",
+	 Warn4("write(%d, %p, "F_Zu"): only wrote "F_Zu" bytes, trying to continue (rev.direction is blocked)",
 	       fd, (const char *)buff+writt, bytes-writt, chk);
 	 writt += chk;
       } else {
@@ -48,6 +52,7 @@ ssize_t writefull(int fd, const void *buff, size_t bytes) {
 	 break;
       }
    }
+   Notice3("write(%d, %p, "F_Zu") completed", fd, (const char *)buff, bytes);
    return writt;
 }
 
@@ -192,9 +197,14 @@ char *sockaddr_info(const struct sockaddr *sa, socklen_t salen, char *buff, size
    switch (sau->soa.sa_family) {
 #if WITH_UNIX
    case 0:
-   case AF_UNIX: sockaddr_unix_info(&sau->un, salen, cp+1, blen-1);
-      cp[0] = '"';
-      strncat(cp+1, "\"", 1);
+   case AF_UNIX:
+      *cp++ = '"';  --blen;
+      sockaddr_unix_info(&sau->un, salen, cp, blen-1);
+      blen -= strlen(cp);
+      cp = strchr(cp, '\0');
+      *cp++ = '"';  --blen;
+      if (blen > 0)
+	 *cp = '\0';
       break;
 #endif
 #if WITH_IP4
@@ -204,6 +214,10 @@ char *sockaddr_info(const struct sockaddr *sa, socklen_t salen, char *buff, size
 #if WITH_IP6
    case AF_INET6: sockaddr_inet6_info(&sau->ip6, cp, blen);
       break;
+#endif
+#if WITH_VSOCK
+   case AF_VSOCK: sockaddr_vm_info(&sau->vm, cp, blen);
+      break;
 #endif
    default:
       n = xio_snprintf(cp, blen, "AF=%d ", sa->sa_family);
@@ -297,6 +311,43 @@ char *sockaddr_inet4_info(const struct sockaddr_in *sa, char *buff, size_t blen)
 }
 #endif /* WITH_IP4 */
 
+#if WITH_VSOCK
+char *sockaddr_vm_info(const struct sockaddr_vm *sa, char *buff, size_t blen) {
+   if (xio_snprintf(buff, blen, "cid:%u port:%u", sa->svm_cid, sa->svm_port) >= blen) {
+      Warn("sockaddr_vm_info(): buffer too short");
+      buff[blen-1] = '\0';
+   }
+   return buff;
+}
+
+int sockaddr_vm_parse(struct sockaddr_vm *sa, const char *cid_str,
+                      const char *port_str)
+{
+   char *garbage = NULL;
+   if (!cid_str) {
+      sa->svm_cid = VMADDR_CID_ANY;
+   } else {
+      sa->svm_cid = strtoul(cid_str, &garbage, 0);
+      if (*garbage != '\0') {
+         Error1("sockaddr_vm - garbage in cid: \"%s\"", garbage);
+         return -EINVAL;
+      }
+   }
+
+   if (!port_str) {
+      sa->svm_port = VMADDR_PORT_ANY;
+   } else {
+      sa->svm_port = strtoul(port_str, &garbage, 0);
+      if (*garbage != '\0') {
+         Error1("sockaddr_vm - garbage in port: \"%s\"", garbage);
+         return -EINVAL;
+      }
+   }
+
+   return 0;
+}
+#endif /* WITH_IP4 */
+
 #if !HAVE_INET_NTOP
 /* http://www.opengroup.org/onlinepubs/000095399/functions/inet_ntop.html */
 const char *inet_ntop(int pf, const void *binaddr,
@@ -416,6 +467,76 @@ char *sockaddr_inet6_info(const struct sockaddr_in6 *sa, char *buff, size_t blen
 }
 #endif /* WITH_IP6 */
 
+
+#if WITH_IP4
+/* Checks if the given string is an IPv4 address.
+   Returns 0 when it is an address.
+   Returns 1 when it is not an address.
+   Returns -1 when an error occurred (bad regex - internal) */
+int check_ip4addr(
+	const char *address)
+{
+#if HAVE_REGEX_H
+	regex_t preg;
+	if (regcomp(&preg, "^(25[0-5]|(2[0-4]|1[0-9]|[1-9]?)[0-9])\\.(25[0-5]|(2[0-4]|1[0-9]|[1-9]?)[0-9])\\.(25[0-5]|(2[0-4]|1[0-9]|[1-9]?)[0-9])\\.(25[0-5]|(2[0-4]|1[0-9]|[1-9]?)[0-9])$", REG_EXTENDED|REG_NOSUB)
+	    != 0) {
+		return -1; 	/* do not handle, just state that no match */
+	}
+	if (regexec(&preg, address, 0, NULL, 0) == 0) {
+		return 0;
+	}
+#endif /* HAVE_REGEX_H */
+	/* Fallback when no regexec() compiled in */
+	return 1;
+}
+#endif /* WITH_IP4 */
+
+#if WITH_IP6
+/* Checks if the given string is an IPv6 address.
+   Currently a hack, just checks a few criteria.
+   Returns 0 when it is an address.
+   Returns 1 when it is not an address.
+   Returns -1 when an error occurred (bad regex - internal) */
+int check_ip6addr(
+	const char *address)
+{
+#if HAVE_REGEX_H
+	regex_t preg;
+	if (regcomp(&preg, "^\\[[0-9a-fA-F:]*\\]$", REG_EXTENDED|REG_NOSUB)
+	    != 0) {
+		return -1; 	/* do not handle, just state that no match */
+	}
+	if (regexec(&preg, address, 0, NULL, 0) == 0) {
+		return 0;
+	}
+#endif /* HAVE_REGEX_H */
+	/* Fallback when no regexec() compiled in */
+	return 1;
+}
+#endif /* WITH_IP6 */
+
+/* Checks if the given string is an IPv6 address.
+   Currently a hack, just checks a few criteria.
+   Returns 0 when it is an address.
+   Returns 1 when it is not an address or on error. */
+int check_ipaddr(
+	const char *address)
+{
+	int res4, res6;
+#if WITH_IP4
+	if ((res4 = check_ip4addr(address)) == 0) {
+		return 0;
+	}
+#endif
+#if WITH_IP6
+	if ((res6 = check_ip6addr(address)) == 0) {
+		return 0;
+	}
+#endif
+	return 1;
+}
+
+
 #if HAVE_GETGROUPLIST || (defined(HAVE_SETGRENT) && defined(HAVE_GETGRENT) && defined(HAVE_ENDGRENT))
 /* fills the list with the supplementary group ids of user.
    caller passes size of list in ngroups, function returns number of groups in
@@ -538,8 +659,8 @@ int xiopoll(struct pollfd fds[], unsigned long nfds, struct timeval *timeout) {
       }
       return result;
    }
-#if HAVE_POLL
    {
+#if HAVE_POLL
       int ms = 0;
       if (timeout == NULL) {
 	 ms = -1;
@@ -549,13 +670,12 @@ int xiopoll(struct pollfd fds[], unsigned long nfds, struct timeval *timeout) {
       /*! timeout */
       return Poll(fds, nfds, ms);
 #else /* HAVE_POLL */
-   } else {
       Error("poll() not available");
       return -1;
 #endif /* !HAVE_POLL */
    }
 }
-   
+
 
 #if WITH_TCP || WITH_UDP
 /* returns port in network byte order;
@@ -705,7 +825,7 @@ int xiosetenv(const char *varname, const char *value, int overwrite, const char
    progname = diag_get_string('p');
    envname[0] = '\0'; strncat(envname, progname, XIO_ENVNAMELEN-1);
    l = strlen(envname);
-   for (i = 0; i < l; ++i)  envname[i] = toupper(envname[i]);
+   for (i = 0; i < l; ++i)  envname[i] = toupper((unsigned char)envname[i]);
    strncat(envname+l, "_", XIO_ENVNAMELEN-l-1);
    l += 1;
    strncat(envname+l, varname, XIO_ENVNAMELEN-l-1);
@@ -731,7 +851,7 @@ int xiosetenv2(const char *varname, const char *varname2, const char *value,
    l += 1;
    strncat(envname+l, varname2, XIO_ENVNAMELEN-l-1);
    l += strlen(envname+l);
-   for (i = 0; i < l; ++i)  envname[i] = toupper(envname[i]);
+   for (i = 0; i < l; ++i)  envname[i] = toupper((unsigned char)envname[i]);
    return _xiosetenv(envname, value, overwrite, sep);
 #  undef XIO_ENVNAMELEN
 }
@@ -759,7 +879,7 @@ int xiosetenv3(const char *varname, const char *varname2, const char *varname3,
    l += 1;
    strncat(envname+l, varname3, XIO_ENVNAMELEN-l-1);
    l += strlen(envname+l);
-   for (i = 0; i < l; ++i)  envname[i] = toupper(envname[i]);
+   for (i = 0; i < l; ++i)  envname[i] = toupper((unsigned char)envname[i]);
    return _xiosetenv(envname, value, overwrite, sep);
 #  undef XIO_ENVNAMELEN
 }
@@ -784,3 +904,250 @@ int xiosetenvushort(const char *varname, unsigned short value, int overwrite) {
    return xiosetenv(varname, envbuff, overwrite, NULL);
 #  undef XIO_SHORTLEN
 }
+
+
+unsigned long int Strtoul(const char *nptr, char **endptr, int base, const char *txt) {
+   unsigned long res;
+
+   res = strtoul(nptr, endptr, base);
+   if (nptr == *endptr) {
+      Error1("parseopts(): missing numerical value of option \"%s\"", txt);
+   }
+   if (**endptr != '\0') {
+      Error1("parseopts(): trailing garbage in numerical arg of option \"%s\"", txt);
+   }
+   return res;
+}
+
+#if HAVE_STRTOLL
+long long int Strtoll(const char *nptr, char **endptr, int base, const char *txt) {
+   long long int res;
+
+   res = strtoul(nptr, endptr, base);
+   if (nptr == *endptr) {
+      Error1("parseopts(): missing numerical value of option \"%s\"", txt);
+   }
+   if (**endptr != '\0') {
+      Error1("parseopts(): trailing garbage in numerical arg of option \"%s\"", txt);
+   }
+   return res;
+}
+#endif /* HAVE_STRTOLL */
+
+double Strtod(const char *nptr, char **endptr, const char *txt) {
+   double res;
+
+   res = strtod(nptr, endptr);
+   if (nptr == *endptr) {
+      Error1("parseopts(): missing numerical value of option \"%s\"", txt);
+   }
+   if (**endptr != '\0') {
+      Error1("parseopts(): trailing garbage in numerical arg of option \"%s\"", txt);
+   }
+   return res;
+}
+
+
+/* This function gets a string with possible references to environment
+   variables and expands them. Variables must be prefixed with '$' and their
+   names consist only of A-Z, a-z, 0-9, and '_'.
+   The name may be enclosed with { }.
+   To pass a literal "$" us "\$".
+   There are special variables supported whose values do not comr from
+   environment:
+   $$ expands to the process's pid
+   $PROGNAME expands to the executables basename or the value of option -lp
+   $TIMESTAMP expands to the actual time in format %Y%m%dT%H%M%S
+   $MICROS expands to the actual microseconds (decimal)
+   The dst output is a copy of src but with the variables expanded.
+   Returns 0 on success;
+   returns -1 when the output buffer was too short (overflow);
+   returns 1 on syntax error.
+*/
+int expandenv(
+	char *dst, 		/* prealloc'd output buff, will be \0 termd */
+	const char *src, 	/* input string to generate expansion from */
+	size_t n, 		/* length of dst */
+	struct timeval *tv) 	/* in/out timestamp for sync */
+{
+	char c; 		/* char currently being lex'd/parsed */
+	bool esc = false;	/* just got '\' */
+	bool bra = false; 	/* within ${ } */
+	char *nam = NULL; 	/* points to temp.allocated rw copy of src */
+	const char *e;		/* pointer to isolated var name in tmp[] */
+	size_t s=0, d=0; 	/* counters in src, dst */
+	size_t v; 		/* variable name begin */
+	bool ofl = false;	/* dst overflow, output truncated */
+	char tmp[18];		/* buffer for timestamp, micros */
+
+	while (c = src[s++]) {
+		if (esc) {
+			if (c == '\0') {
+				if (d+2 > n)  { ofl = true; break; }
+				dst[d++] = '\\';
+				dst[d++] = c;
+				break;
+			}
+			if (c != '$') {
+				if (d+3 > n)  { ofl = true; break; }
+				dst[d++] = '\\';
+			} else {
+				if (d+2 > n)  { ofl = true; break; }
+			}
+			dst[d++] = c;
+			esc = false;
+			continue;
+		}
+		if (c == '\0') {
+			dst[d++] = c;
+			break;
+		}
+		if (c == '\\') {
+			esc = true;
+			continue;
+		}
+		if (c != '$') {
+			if (d+2 > n)  { ofl = true; break; }
+			dst[d++] = c;
+			continue;
+		}
+
+		/* c == '$': Expecting env var to expand */
+		c = src[s];
+		if (c == '\0') {
+			if (d+2 > n)  { ofl = true; break; }
+			++s;
+			dst[d++] = '$';
+			break;
+		}
+		if (c == '$') {
+			int wr;
+			/* Special case: pid */
+			++s;
+			wr = snprintf(&dst[d], n-d, F_pid, getpid());
+			if (wr >= n-d || wr < 0)  { ofl = true; break; }
+			d += wr;
+			continue;
+		}
+		if (c == '{') {
+			++s;
+			bra = true;
+		}
+		if (!isalpha(c) && c != '_') {
+			/* Special case no var name, just keep '$' */
+			if (d+3 > n)  { ofl = true; break; }
+			++s;
+			dst[d++] = '$';
+			dst[d++] = c;
+			continue;
+		}
+		v = 0; 		/* seems we found valid variable name */
+		if (nam == NULL) {
+			nam = strdup(src+s);
+			if (nam == NULL) {
+				errno = ENOMEM;
+				return -1;
+			}
+		}
+		while (c = src[s]) {
+			if (!isalnum(c) && c != '_') {
+				break;
+			}
+			nam[v++] = c;
+			++s;
+		}
+		if (bra && c != '}') {
+			return 1;
+		}
+
+		nam[v] = '\0';
+		/* Var name is complete */
+		/* Check hardcoded var names */
+		if (strcmp(nam, "PROGNAME") == 0) {
+			e = diag_get_string('p');
+		} else if (strcmp(nam, "TIMESTAMP") == 0) {
+			if (tv->tv_sec == 0) {
+				gettimeofday(tv, NULL);
+			}
+			struct tm tm;
+			localtime_r(&tv->tv_sec, &tm);
+			strftime(tmp, sizeof(tmp), "%Y%m%dT%H%M%S", &tm);
+			e = tmp;
+		} else if (strcmp(nam, "MICROS") == 0) {
+			if (tv->tv_sec == 0) {
+				gettimeofday(tv, NULL);
+			}
+			sprintf(tmp, F_tv_usec, tv->tv_usec);
+			e = tmp;
+		} else {
+			e = getenv(nam);
+		}
+		if (e == NULL) {
+			/* Var not found, skip it */
+			continue;
+		}
+		/* Var found, copy it to output buffer */
+		if (d+strlen(e)+1 > n)  { ofl = true; break; }
+		strcpy(&dst[d], e);
+		d += strlen(&dst[d]);
+		continue;
+	}
+
+	if (nam != NULL) {
+		free(nam);
+	}
+
+	if (ofl) {
+		dst[d] = '\0';
+		return -1;
+	}
+	dst[d] = '\0';
+	if (src[s-1] != '\0') {
+		errno = EINVAL;
+		return -1;
+	}
+	return 0;
+}
+
+int xio_opensnifffile(
+	const char *a,
+	struct timeval *tv)
+{
+	char path[PATH_MAX];
+	int flags;
+	int rc;
+	int fd;
+
+	rc = expandenv(path, a, sizeof(path), tv);
+	if (rc < 0) {
+		Error2("expandenv(source=\"%s\", n="F_Zu"): Out of memory", a, sizeof(path));
+		errno = ENOMEM;
+		return -1;
+	} else if (rc > 0) {
+		Error1("expandenv(source=\"%s\"): Syntax error", a);
+		errno = EINVAL;
+		return -1;
+	}
+	flags = O_CREAT|O_WRONLY|O_APPEND|
+#ifdef O_LARGEFILE
+		O_LARGEFILE|
+#endif
+#ifdef O_CLOEXEC
+		O_CLOEXEC|
+#endif
+		O_NONBLOCK;
+	if ((fd = Open(path, flags, 0664)) < 0) {
+		if (errno == ENXIO) {
+			/* try to open pipe rdwr */
+			if ((fd = Open(path, flags, 0664)) < 0) {
+				return -1;
+			}
+		}
+	}
+#ifdef O_CLOEXEC
+	if (Fcntl_l(fd, F_SETFD, FD_CLOEXEC) < 0) {
+		Warn2("fcntl(%d, F_SETFD, FD_CLOEXEC): %s", fd, strerror(errno));
+	}
+#endif
+	return fd;
+}
