@@ -1,5 +1,5 @@
 /* source: utils.c */
-/* Copyright Gerhard Rieger */
+/* Copyright Gerhard Rieger and contributors (see file CHANGES) */
 /* Published under the GNU General Public License V.2, see file COPYING */
 
 /* useful additions to C library */
@@ -67,7 +67,7 @@ const struct wordent *keyw(const struct wordent *keywds, const char *name, unsig
    return NULL;
 }
 
-/* Linux: setenv(), AIX: putenv() */
+/* Linux: setenv(), AIX (4.3?): putenv() */
 #if !HAVE_SETENV
 int setenv(const char *name, const char *value, int overwrite) {
    int result;
@@ -89,7 +89,6 @@ int setenv(const char *name, const char *value, int overwrite) {
 #endif /* !HAVE_SETENV */
 
 
-
 /* sanitizes an "untrusted" character. output buffer must provide at least 4
    characters space.
    Does not append \0. returns length of output (currently: max 4) */
@@ -97,7 +96,7 @@ static size_t sanitize_char(char c, char *o, int style) {
    int hn;	/* high nibble */
    int ln;	/* low nibble */
    int n;	/* written chars */
-   if (isprint(c)) {
+   if (isprint((unsigned char)c)) {
       *o = c;
       return 1;
    }
@@ -160,7 +159,7 @@ char *xiosubstr(char *scratch, const char *str, size_t from, size_t len) {
    *scratch = '\0';
    return scratch0;
 }
-      
+
 
 /* since version 1.7.2.4 socat supports C-99 behaviour of snprintf but still
    can handle the old glibc case with -1 return on truncation.
