@@ -1,5 +1,5 @@
 /* source: xio-file.h */
-/* Copyright Gerhard Rieger 2001-2006 */
+/* Copyright Gerhard Rieger and contributors (see file CHANGES) */
 /* Published under the GNU General Public License V.2, see file COPYING */
 
 #ifndef __xio_file_h_included
@@ -26,6 +26,6 @@ extern const struct optdesc opt_o_priv;
 extern const struct optdesc opt_o_trunc;
 extern const struct optdesc opt_o_noatime;
 
-extern const struct addrdesc addr_open;
+extern const struct addrdesc xioaddr_open;
 
 #endif /* !defined(__xio_file_h_included) */
