@@ -1,11 +1,11 @@
 /* source: xio-readline.h */
-/* Copyright Gerhard Rieger 2002, 2003 */
+/* Copyright Gerhard Rieger and contributors (see file CHANGES) */
 /* Published under the GNU General Public License V.2, see file COPYING */
 
 #ifndef __xio_readline_h_included
 #define __xio_readline_h_included 1
 
-extern const struct addrdesc addr_readline;
+extern const struct addrdesc xioaddr_readline;
 
 extern const struct optdesc opt_history_file;
 extern const struct optdesc opt_prompt;
