@@ -44,6 +44,8 @@ extern const struct optdesc opt_so_type;
 extern const struct optdesc opt_so_dontroute;
 extern const struct optdesc opt_so_rcvlowat;
 extern const struct optdesc opt_so_sndlowat;
+extern const struct optdesc opt_so_rcvtimeo;
+extern const struct optdesc opt_so_sndtimeo;
 extern const struct optdesc opt_so_audit;
 extern const struct optdesc opt_so_attach_filter;
 extern const struct optdesc opt_so_detach_filter;
@@ -82,6 +84,7 @@ extern
 char *xiogetifname(int ind, char *val, int ins);
 
 extern int retropt_socket_pf(struct opt *opts, int *pf);
+extern int xiogetancillary(int fd, struct msghdr *msgh, int flags);
 
 extern int xioopen_connect(struct single *fd,
 			    union sockaddr_union *us, size_t uslen,
@@ -97,11 +100,11 @@ extern int _xioopen_connect(struct single *fd,
 			    bool alt, int level);
 
 /* common to xioopen_udp_sendto, ..unix_sendto, ..rawip */
-extern 
+extern
 int _xioopen_dgram_sendto(/* them is already in xfd->peersa */
 			union sockaddr_union *us, socklen_t uslen,
 			struct opt *opts,
-			int xioflags, xiosingle_t *xfd, unsigned groups,
+			int xioflags, xiosingle_t *xfd, groups_t groups,
 			int pf, int socktype, int ipproto, bool alt);
 extern
 int _xioopen_dgram_recvfrom(struct single *xfd, int xioflags,
@@ -113,9 +116,8 @@ int _xioopen_dgram_recv(struct single *xfd, int xioflags,
 			struct sockaddr *us, socklen_t uslen,
 			struct opt *opts, int pf, int socktype, int proto,
 			int level);
+extern int xiodopacketinfo(struct single *sfd, struct msghdr *msgh, bool withlog, bool withenv);
 extern
-int xiodopacketinfo(struct msghdr *msgh, bool withlog, bool withenv);
-extern 
 int xiogetpacketsrc(int fd, struct msghdr *msgh, int flags);
 extern
 int xiocheckpeer(xiosingle_t *xfd,
@@ -123,15 +125,13 @@ int xiocheckpeer(xiosingle_t *xfd,
 extern
 int xiosetsockaddrenv(const char *lr, union sockaddr_union *sau, socklen_t salen, int proto);
 
-extern
-int xioparsenetwork(const char *rangename, int pf,
-		    struct xiorange *range);
-extern 
-int xioparserange(const char *rangename, int pf, struct xiorange *range);
+extern int xioparsenetwork(const char *rangename, int pf, struct xiorange *range, const int ai_flags[2]);
+extern int xioparserange(const char *rangename, int pf, struct xiorange *range, const int ai_flags[2]);
 
 extern int
 xiosocket(struct opt *opts, int pf, int socktype, int proto, int level);
-extern int 
+extern int
 xiosocketpair(struct opt *opts, int pf, int socktype, int proto, int sv[2]);
+extern int xiosock_reuseaddr(int fd, int ipproto, struct opt *opts);
 
 #endif /* !defined(__xio_socket_h_included) */
