@@ -1,5 +1,5 @@
 /* source: xiohelp.c */
-/* Copyright Gerhard Rieger 2001-2009 */
+/* Copyright Gerhard Rieger and contributors (see file CHANGES) */
 /* Published under the GNU General Public License V.2, see file COPYING */
 
 /* this file contains the source for the help function */
@@ -14,15 +14,15 @@
 /* keep consistent with xioopts.h:enum e_types ! */
 static const char *optiontypenames[] = {
 	"CONST",	"BIN",		"BOOL",		"BYTE",
-	"INT",		"LONG",		"STRING",	"PTRDIFF",
+	"INT",		"INT/NULL",	"LONG",		"STRING",	"PTRDIFF",
 	"SHORT",	"SIZE_T",	"SOCKADDR",	"UNSIGNED-INT",
 	"UNSIGNED-LONG","UNSIGNED-SHORT","MODE_T",	"GID_T",
 	"UID_T",	"INT[3]",	"STRUCT-TIMEVAL", "STRUCT-TIMESPEC",
 	"DOUBLE",	"STRING-NULL",	"LONG-LONG",	"OFF_T",
 	"OFF64_T",	"INT:INT",	"INT:INTP",	"INT:BIN",
 	"INT:STRING",	"INT:INT:INT",	"INT:INT:BIN",	"INT:INT:STRING",
-	"IP4NAME",
-
+	"INT:INT:GENERIC",
+	"IP4NAME",	"IP4SOCK",
 #if HAVE_STRUCT_LINGER
 			"STRUCT-LINGER",
 #endif
@@ -31,6 +31,10 @@ static const char *optiontypenames[] = {
 #elif HAVE_STRUCT_IP_MREQ
 					"STRUCT-IP_MREQ",
 #endif
+#if HAVE_STRUCT_IP_MREQ_SOURCE
+	"IP-MREQ-SOURCE",
+#endif
+	"GENERIC",
 } ;
 
 
@@ -39,14 +43,14 @@ static const char *addressgroupnames[] = {
 	"FD",		"FIFO",		"CHR",		"BLK",
 	"REG",		"SOCKET",	"READLINE",	"undef",
 	"NAMED",	"OPEN",		"EXEC",		"FORK",
-	"LISTEN",	"DEVICE",	"CHILD",	"RETRY",
+	"LISTEN",	"SHELL",	"CHILD",	"RETRY",
 	"TERMIOS",	"RANGE",	"PTY",		"PARENT",
 	"UNIX",		"IP4",		"IP6",		"INTERFACE",
 	"UDP",		"TCP",		"SOCKS4",	"OPENSSL",
-	"PROCESS",	"APPL",		"HTTP",		"SCTP"
+	"PROCESS",	"APPL",		"HTTP",		"undef",
+	"POSIXMQ",	"SCTP",		"DCCP",		"UDPLITE"
 } ;
 
-
 /* keep consistent with xioopts.h:enum ephase ! */
 static char *optionphasenames[] = {
 	"ALL",		"INIT",		"EARLY",
@@ -54,57 +58,80 @@ static char *optionphasenames[] = {
 	"PRESOCKET",	"SOCKET",	"PASTSOCKET",
 	"PREBIGEN",	"BIGEN",	"PASTBIGEN",
 	"FD",
-	"PREBIND",	"BIND",		"PASTBIND",	
+	"PREBIND",	"BIND",		"PASTBIND",
 	"PRELISTEN",	"LISTEN",	"PASTLISTEN",
 	"PRECONNECT",	"CONNECT",	"PASTCONNECT",
 	"PREACCEPT",	"ACCEPT",	"PASTACCEPT",
 	"CONNECTED",
 	"PREFORK",	"FORK",		"PASTFORK",
 	"LATE",		"LATE2",
-	"PREEXEC",	"EXEC",		"SPECIFIC",
+	"PREEXEC",	"EXEC",		"PASTEXEC",
+	"SPECIFIC",
 	NULL
 } ;
 
 
 /* print a line about a single option */
 static int xiohelp_option(FILE *of, const struct optname *on, const char *name) {
-   int j;
-   unsigned int groups;
+   int chars;
+   int i, j;
+   groups_t groups;
    bool occurred;
 
-   fprintf(of, "      %s\tgroups=", name);
-   groups = on->desc->group;  occurred = false;
-   for (j = 0; j < 32; ++j) {
+   chars = fprintf(of, "      %s", name);
+   i = (16 - chars + 7) / 8;
+   for (; i > 0; --i) { fputc('\t', of); }
+   fputc('\t', of);
+
+   fputs("groups=", of);
+   groups = on->desc->group;
+   occurred = false;
+   chars = 7;
+   for (j = 0; j < 8*sizeof(groups_t); ++j) {
       if (groups & 1) {
-	 if (occurred)  { fputc(',', of); }
-	 fprintf(of, "%s", addressgroupnames[j]);
+	 if (occurred) {
+	    fputc(',', of);
+	    ++chars;
+	 }
+	 fputs(addressgroupnames[j], of);
+	 chars += strlen(addressgroupnames[j]);
 	 occurred = true;
       }
       groups >>= 1;
    }
-   fprintf(of, "\tphase=%s", optionphasenames[on->desc->phase]);
-   fprintf(of, "\ttype=%s", optiontypenames[on->desc->type]);
+   i = (24 - chars + 7) / 8;
+   for (; i > 0; --i) { fputc('\t', of); }
+
+   chars = fprintf(of, "phase=%s", optionphasenames[on->desc->phase]);
+   i = (24 - chars + 7) / 8;
+   for (; i > 0; --i) { fputc('\t', of); }
+
+   fprintf(of, "type=%s", optiontypenames[on->desc->type]);
    fputc('\n', of);
    return 0;
 }
 
+const char *xiohelp_opttypename(enum e_types typnum)
+{
+	if (typnum < 0 || typnum >= TYPE_OVERFLOW) {
+		Warn2("%s(): invalid type number %d", __func__, typnum);
+		return "<invalid>";
+	}
+	return optiontypenames[typnum];
+}
+
 int xioopenhelp(FILE *of,
 	       int level	/* 0..only addresses, 1..and options */
 	       ) {
    const struct addrname *an;
    const struct optname *on;
    int i, j;
-   unsigned int groups;
+   groups_t groups;
    bool occurred;
 
-   fputs("   bi-address:\n", of);
-   fputs("      pipe[,<opts>]\tgroups=FD,FIFO\n", of);
-   if (level == 2) {
-      fputs("      echo is an alias for pipe\n", of);
-      fputs("      fifo is an alias for pipe\n", of);
-   }
-   fputs("      <single-address>!!<single-address>\n", of);
+   fputs("   bi-address:  /* is an address that may act both as data sync and source */\n", of);
    fputs("      <single-address>\n", of);
+   fputs("      <single-address>!!<single-address>\n", of);
    fputs("   single-address:\n", of);
    fputs("      <address-head>[,<opts>]\n", of);
    fputs("   address-head:\n", of);
@@ -112,10 +139,16 @@ int xioopenhelp(FILE *of,
    i = 0;
    while (addressnames[i].name) {
       if (!strcmp(an->name, an->desc->defname)) {
+	 int chars, i;
+
          /* it is a canonical address name */
-	 fprintf(of, "      %s", an->name);
+	 chars = fprintf(of, "      %s", an->name);
 	 if (an->desc->syntax) {
-	    fputs(an->desc->syntax, of); }
+	    fputs(an->desc->syntax, of);
+	    chars += strlen(an->desc->syntax);
+	 }
+	 i = (40 - chars + 7) / 8;
+	 for (; i > 0; --i) { fputc('\t', of); }
 	 fputs("\tgroups=", of);
 	 groups = an->desc->groups;  occurred = false;
 	 for (j = 0; j < 32; ++j) {
@@ -128,7 +161,13 @@ int xioopenhelp(FILE *of,
 	 }
 	 fputc('\n', of);
       } else if (level == 2) {
-         fprintf(of, "         %s is an alias name for %s\n", an->name, an->desc->defname);
+	 int chars, i;
+
+         chars = fprintf(of, "      %s", an->name);
+	 i = (40 - chars + 7) / 8;
+	 for (; i > 0; --i) { fputc('\t', of); }
+
+         fprintf(of, "\tis an alias name for %s\n", an->desc->defname);
       }
       ++an; ++i;
    }
@@ -144,10 +183,16 @@ int xioopenhelp(FILE *of,
    fputs("   opt:\n", of);
    on = optionnames;
    while (on->name != NULL) {
-      if (on->desc->nickname!= NULL
+      if (on->desc->nickname != NULL
 	  && !strcmp(on->name, on->desc->nickname)) {
 	 if (level == 2) {
-	    fprintf(of, "      %s is an alias for %s\n", on->name, on->desc->defname);
+	    int chars, i;
+
+	    chars = fprintf(of, "      %s", on->name);
+	    i = (16 - chars + 7) / 8;
+	    for (; i > 0; --i) { fputc('\t', of); }
+
+	    fprintf(of, "\tis an alias for %s\n", on->desc->defname);
 	 } else {
 	    xiohelp_option(of, on, on->name);
 	 }
@@ -158,7 +203,13 @@ int xioopenhelp(FILE *of,
 	 if (!strcmp(on->name, on->desc->defname)) {
 	    xiohelp_option(of, on, on->name);
 	 } else {
-	    fprintf(of, "      %s is an alias for %s\n", on->name, on->desc->defname);
+	    int chars, i;
+
+	    chars = fprintf(of, "      %s", on->name);
+	    i = (16 - chars + 7) / 8;
+	    for (; i > 0; --i) { fputc('\t', of); }
+
+	    fprintf(of, "\tis an alias for %s\n", on->desc->defname);
 	 }
       }
       ++on;
@@ -167,4 +218,16 @@ int xioopenhelp(FILE *of,
    return 0;
 }
 
+/* This function may be used by address handling code to log syntax error */
+int xiohelp_syntax(
+	const char *addr,
+	int expectnum,
+	int isnum,
+	const char *syntax)
+{
+	Error5("%s: wrong number of parameters (%d instead of %d): usage: %s%s",
+	       addr, isnum, expectnum, addr, syntax);
+	return -1;
+}
+
 #endif /* WITH_HELP */
