@@ -15,6 +15,7 @@
 #include "xio-creat.h"
 #include "xio-gopen.h"
 #include "xio-pipe.h"
+#include "xio-socketpair.h"
 #if _WITH_SOCKET
 #include "xio-socket.h"
 #include "xio-listen.h"
@@ -30,18 +31,25 @@
 #include "xio-tcp.h"
 #include "xio-udp.h"
 #include "xio-sctp.h"
+#include "xio-dccp.h"
+#include "xio-udplite.h"
 #include "xio-socks.h"
+#include "xio-socks5.h"
 #include "xio-proxy.h"
+#include "xio-vsock.h"
 #endif /* _WITH_SOCKET */
+#include "xio-namespaces.h"
+#include "xio-posixmq.h"
 #include "xio-progcall.h"
 #include "xio-exec.h"
 #include "xio-system.h"
+#include "xio-shell.h"
 #include "xio-termios.h"
 #include "xio-readline.h"
 #include "xio-pty.h"
 #include "xio-openssl.h"
 #include "xio-tcpwrap.h"
-#include "xio-ext2.h"
+#include "xio-fs.h"
 #include "xio-tun.h"
 #include "xio-streams.h"
 
