@@ -1,5 +1,5 @@
 /* source: xioopts.h */
-/* Copyright Gerhard Rieger 2001-2009 */
+/* Copyright Gerhard Rieger and contributors (see file CHANGES) */
 /* Published under the GNU General Public License V.2, see file COPYING */
 
 #ifndef __xioopts_h_included
@@ -27,6 +27,7 @@ enum e_types {
    TYPE_BYTE,		/* unsigned char */
 
    TYPE_INT,		/* int */
+   TYPE_INT_NULL,	/* int, or just opt= for no action (instead of default) */
    TYPE_LONG,		/* long */
    TYPE_STRING,		/* char * */
    TYPE_NAME = TYPE_STRING,
@@ -64,18 +65,30 @@ enum e_types {
    TYPE_INT_INT_INT,	/* 3 params: first and second are int, 3rd is int */
    TYPE_INT_INT_BIN,	/* 3 params: first and second are int, 3rd is binary */
    TYPE_INT_INT_STRING,	/* 3 params: first and second are int, 3rd is string */
+   TYPE_INT_INT_GENERIC,	/* 3 params: first and second are int, 3rd is specified by value (dalan syntax) */
 
    TYPE_IP4NAME,	/* IPv4 hostname or address */
+   TYPE_IP4SOCK,	/* IPv4 hostname or address optionally with port */
 #if HAVE_STRUCT_LINGER
    TYPE_LINGER,		/* struct linger */
 #endif /* HAVE_STRUCT_LINGER */
 #if HAVE_STRUCT_IP_MREQ || HAVE_STRUCT_IP_MREQN
    TYPE_IP_MREQN,	/* for  struct ip_mreq  or  struct ip_mreqn */
 #endif
+#if HAVE_STRUCT_IP_MREQ_SOURCE
+   TYPE_IP_MREQ_SOURCE,	/* for  struct ip_mreq_source */
+#endif
+#if HAVE_STRUCT_GROUP_SOURCE_REQ
+   TYPE_GROUP_SOURCE_REQ,	/* for  struct group_source_req */
+#endif
+
+   TYPE_GENERIC, 	/* type is determined from (text) data provided (dalan syntax) */
+   TYPE_OVERFLOW 	/* marker: beyond last type */
 } ;
 
 enum e_func {
    OFUNC_NONE,		/* no function - should not occur */
+   /* Options to be applied to FD directly */
    OFUNC_FLAG,		/* no function, but bitposition, only with bool; arg1 is mask */
    OFUNC_FLAG_PATTERN,	/* no function, but bitpattern: arg1 is pattern, arg2 is mask */
    OFUNC_SEEK32,	/* lseek(): arg1 is whence (SEEK_SET etc.) */
@@ -90,12 +103,6 @@ enum e_func {
    OFUNC_SOCKOPT_APPEND,/* getsockopt(), append data, setsockopt() */
    OFUNC_SOCKOPT_GENERIC,/* generic setsockopt() (level, optname on cmdline) */
    OFUNC_FLOCK,		/* flock() */
-   OFUNC_TERMIO,	/* termio() ? */
-   OFUNC_SPEC,		/* special, i.e. no generalizable function call */
-   OFUNC_OFFSET,	/* put a value into xiofile struct; major is offset */
-   OFUNC_OFFSET_MASKS,	/* !!! */
-   /*OFUNC_APPL,*/	/* special, i.e. application must know which f. */
-   OFUNC_EXT,		/* with extended file descriptors only */
    OFUNC_TERMIOS_FLAG,	/* a flag in struct termios: major..tcflag, minor..bit
 			 */
    OFUNC_TERMIOS_PATTERN, /* a multibit: major..tcflag, minor..pattern,
@@ -104,20 +111,31 @@ enum e_func {
    OFUNC_TERMIOS_CHAR,	/* a termios functional character: major..c_cc index */
    OFUNC_TERMIOS_SPEED,	/* termios c_ispeed etc on FreeBSD */
    OFUNC_TERMIOS_SPEC,	/* termios combined modes */
+
+   OFUNC_EXT,		/* with extended file descriptors only */
+#  define OFUNC_XIO OFUNC_EXT
+   OFUNC_SPEC,		/* special, i.e. no generalizable function call */
+   OFUNC_OFFSET,	/* put a value into xiofile struct; major is offset */
+   OFUNC_OFFSET_MASK,	/* set/unset bit pattern in variable */
+   OFUNC_OFFSET_MASKS,	/* set pos or neg bit pattern in array[2] */
+   /*OFUNC_APPL,*/	/* special, i.e. application must know which f. */
    OFUNC_SIGNAL,	/* a signal that should be passed to child process */
    OFUNC_RESOLVER,	/* a bit position used on _res.options */
-   OFUNC_IFFLAG,	/* interface flag: locical-or a 1bit mask */
+   OFUNC_IFFLAG,	/* interface flag: logical-or a 1bit mask */
+   OFUNC_SET_NAMESPACE, 	/* set/change Linux namespace */
+   OFUNC_RESET_NAMESPACE,	/* set Linux namespace back to default */
 #  define ENABLE_OFUNC
 #  include "xio-streams.h"	/* push a POSIX STREAMS module */
 #  undef ENABLE_OFUNC
+   OFUNC_OVERFLOW 	/* marker: beyond last func */
 } ;
 
 /* for simpler handling of option-to-connection-type relations we define
    groups. to keep the search for options simple, we allow each option to
-   belong to at most one group only. (we have a dummy GROUP_NONE for those 
+   belong to at most one group only. (we have a dummy GROUP_NONE for those
    that don't want to belong to any...)
-   The caller of parseopts() specifies per bitpatter a set of groups where it
-   accepts options from. 
+   The caller of parseopts() specifies per bit pattern a set of groups where it
+   accepts options from.
 */
 
 /*- the group bits are:
@@ -130,10 +148,11 @@ enum e_func {
 /* keep consistent with xiohelp.c:addressgroupnames[] ! */
 /* a dummy group */
 #define GROUP_NONE	0x00000000
+#define GROUP_ADDR	0x00000000 	/* options that apply to all addresses */
 
 #define GROUP_FD	0x00000001	/* everything applyable to a fd */
 #define GROUP_FIFO	0x00000002
-#define GROUP_CHR	0x00000004
+#define GROUP_CHR	0x00000004 	/* not yet used? */
 #define GROUP_BLK	0x00000008
 #define GROUP_REG	0x00000010
 #define GROUP_FILE GROUP_REG
@@ -146,7 +165,7 @@ enum e_func {
 #define GROUP_FORK	0x00000800	/* communication with forked process */
 
 #define GROUP_LISTEN	0x00001000	/* socket in listening mode */
-/*	0x00002000 */
+#define GROUP_SHELL	0x00002000
 #define GROUP_CHILD	0x00004000	/* autonom child process */
 #define GROUP_RETRY	0x00008000	/* when open/connect etc. fails */
 #define GROUP_TERMIOS	0x00010000
@@ -155,25 +174,39 @@ enum e_func {
 #define GROUP_PARENT	0x00080000	/* for parent of communicating child */
 
 #define GROUP_SOCK_UNIX	0x00100000
-#define GROUP_SOCK_IP4	0x00200000
+#define GROUP_SOCK_IP4	0x00200000 	/* not yet used? */
 #define GROUP_SOCK_IP6	0x00400000
 #define GROUP_SOCK_IP	(GROUP_SOCK_IP4|GROUP_SOCK_IP6)
 #define GROUP_INTERFACE	0x00800000
 #define GROUP_TUN       GROUP_INTERFACE
 
-#define GROUP_IP_UDP	0x01000000
+#define GROUP_IP_UDP	0x01000000 	/* not yet used? */
 #define GROUP_IP_TCP	0x02000000
-#define GROUP_IPAPP	(GROUP_IP_UDP|GROUP_IP_TCP|GROUP_IP_SCTP)	/* true: indicates one of UDP, TCP, SCTP */
 #define GROUP_IP_SOCKS4	0x04000000
 #define GROUP_OPENSSL	0x08000000
 
 #define GROUP_PROCESS	0x10000000	/* a process related option */
 #define GROUP_APPL	0x20000000	/* option handled by data loop */
 #define GROUP_HTTP	0x40000000	/* any HTTP client */
-#define GROUP_IP_SCTP	0x80000000
 
+/* Keep condition consistent with xio.h:groups_t! */
+#if WITH_POSIXMQ || WITH_SCTP || WITH_DCCP || WITH_UDPLITE
+/* The following groups are not expected on systems without uint64_t */
+#define GROUP_POSIXMQ	((groups_t)0x0100000000ULL)
+#define GROUP_IP_SCTP	((groups_t)0x0200000000ULL)
+#define GROUP_IP_DCCP	((groups_t)0x0400000000ULL)
+#define GROUP_IP_UDPLITE ((groups_t)0x0800000000ULL)
+#define GROUP_ALL	((groups_t)0x0fffffffffULL) 	/* OMG... for 32bit systems */
+#else /* !(WITH_POSIXMQ || WITH_SCTP || WITH_DCCP || WITH_UDPLITE) */
+#define GROUP_POSIXMQ	0
+#define GROUP_IP_SCTP	0
+#define GROUP_IP_DCCP	0
+#define GROUP_IP_UDPLITE	0
+#define GROUP_ALL	0xffffffffU
+#endif /* !(WITH_POSIXMQ || WITH_SCTP || WITH_DCCP || WITH_UDPLITE) */
+
+#define GROUP_IPAPP	(GROUP_IP_UDP|GROUP_IP_TCP|GROUP_IP_SCTP|GROUP_IP_DCCP|GROUP_IP_UDPLITE)	/* true: indicates one of UDP, TCP, SCTP, DCCP, UDPlite */
 #define GROUP_ANY	(GROUP_PROCESS|GROUP_APPL)
-#define GROUP_ALL	0xffffffff
 
 
 /* no IP multicasts, no error queue yet */
@@ -182,6 +215,9 @@ enum e_func {
 /* optcode's */
 enum e_optcode {
    OPT_ADDRESS_FAMILY = 1,
+   OPT_AI_ADDRCONFIG, 	/* getaddrinfo() */
+   OPT_AI_PASSIVE, 	/* getaddrinfo() */
+   OPT_AI_V4MAPPED, 	/* getaddrinfo() */
    /* these are not alphabetically, I know... */
    OPT_B0,		/* termios.c_cflag */
    OPT_B50,		/* termios.c_cflag */
@@ -229,6 +265,8 @@ enum e_optcode {
 #  endif
    OPT_BSDLY,		/* termios.c_oflag */
 #endif
+   OPT_CHDIR, 		/* change working directory */
+   OPT_CHILDREN_SHUTUP,
    OPT_CHROOT,		/* chroot() past file system access */
    OPT_CHROOT_EARLY,	/* chroot() before file system access */
    /*OPT_CIBAUD,*/		/* termios.c_cflag */
@@ -264,6 +302,7 @@ enum e_optcode {
    OPT_CSIZE,		/* termios.c_cflag */
    OPT_CSTOPB,		/* termios.c_cflag */
    OPT_DASH,		/* exec() */
+   OPT_DCCP_SET_CCID,
    OPT_ECHO,		/* termios.c_lflag */
    OPT_ECHOCTL,		/* termios.c_lflag */
    OPT_ECHOE,		/* termios.c_lflag */
@@ -275,18 +314,6 @@ enum e_optcode {
 #endif
    OPT_END_CLOSE,	/* xfd.stream.howtoend = END_CLOSE */
    OPT_ESCAPE,
-   OPT_EXT2_SECRM,
-   OPT_EXT2_UNRM,
-   OPT_EXT2_COMPR,
-   OPT_EXT2_SYNC,
-   OPT_EXT2_IMMUTABLE,
-   OPT_EXT2_APPEND,
-   OPT_EXT2_NODUMP,
-   OPT_EXT2_NOATIME,
-   OPT_EXT2_JOURNAL_DATA,
-   OPT_EXT2_NOTAIL,
-   OPT_EXT2_DIRSYNC,
-   OPT_EXT2_TOPDIR,
    OPT_FDIN,
    OPT_FDOUT,
 #ifdef FFDLY
@@ -309,16 +336,30 @@ enum e_optcode {
    /*0 OPT_FORCE,*/
    OPT_FOREVER,
    OPT_FORK,
+   OPT_FS_APPEND,
+   OPT_FS_COMPR,
+   OPT_FS_DIRSYNC,
+   OPT_FS_IMMUTABLE,
+   OPT_FS_JOURNAL_DATA,
+   OPT_FS_NOATIME,
+   OPT_FS_NODUMP,
+   OPT_FS_NOTAIL,
+   OPT_FS_SECRM,
+   OPT_FS_SYNC,
+   OPT_FS_TOPDIR,
+   OPT_FS_UNRM,
    OPT_FTRUNCATE32,	/* ftruncate() */
    OPT_FTRUNCATE64,	/* ftruncate64() */
    OPT_F_SETLKW_RD,	/* fcntl with struct flock - read-lock, wait */
    OPT_F_SETLKW_WR,	/* fcntl with struct flock - write-lock, wait */
    OPT_F_SETLK_RD,	/* fcntl with struct flock - read-lock */
    OPT_F_SETLK_WR,	/* fcntl with struct flock - write-lock */
+   OPT_F_SETPIPE_SZ, 	/* pipe == fifo */
    OPT_GROUP,
    OPT_GROUP_EARLY,
    OPT_GROUP_LATE,
    OPT_HISTORY_FILE,	/* readline history file */
+   OPT_HTTP_VERSION,
    OPT_HUPCL,		/* termios.c_cflag */
    OPT_ICANON,		/* termios.c_lflag */
    OPT_ICRNL,		/* termios.c_iflag */
@@ -355,6 +396,7 @@ enum e_optcode {
    OPT_IPV6_HOPLIMIT,
    OPT_IPV6_HOPOPTS,
    OPT_IPV6_JOIN_GROUP,
+   OPT_IPV6_JOIN_SOURCE_GROUP,
    OPT_IPV6_PKTINFO,
    OPT_IPV6_RECVDSTOPTS,
    OPT_IPV6_RECVERR,
@@ -377,6 +419,7 @@ enum e_optcode {
    OPT_IOCTL_STRING,	/* generic ioctl with integer value (pointed to) */
    OPT_IOCTL_VOID,	/* generic ioctl without value */
    OPT_IP_ADD_MEMBERSHIP,
+   OPT_IP_ADD_SOURCE_MEMBERSHIP,
 #ifdef IP_HDRINCL
    OPT_IP_HDRINCL,
 #endif
@@ -420,6 +463,9 @@ enum e_optcode {
    OPT_IP_ROUTER_ALERT,
 #endif
    OPT_IP_TOS,
+#ifdef IP_TRANSPARENT
+   OPT_IP_TRANSPARENT,
+#endif
    OPT_IP_TTL,
    OPT_ISIG,		/* termios.c_lflag */
    OPT_ISPEED,		/* termios.c_ispeed */
@@ -430,9 +476,13 @@ enum e_optcode {
    OPT_IXANY,		/* termios.c_iflag */
    OPT_IXOFF,		/* termios.c_iflag */
    OPT_IXON,		/* termios.c_iflag */
+   OPT_ACCEPT_TIMEOUT,	/* listening socket */
    OPT_LOCKFILE,
    OPT_LOWPORT,
    OPT_MAX_CHILDREN,
+#if WITH_POSIXMQ
+   OPT_POSIXMQ_PRIORITY,
+#endif
 #ifdef NLDLY
 #  ifdef NL0
    OPT_NL0,		/* termios.c_oflag */
@@ -473,6 +523,7 @@ enum e_optcode {
    OPT_OPENSSL_CAPATH,
    OPT_OPENSSL_CERTIFICATE,
    OPT_OPENSSL_CIPHERLIST,
+   OPT_OPENSSL_COMMONNAME,
 #if OPENSSL_VERSION_NUMBER >= 0x00908000L
    OPT_OPENSSL_COMPRESS,
 #endif
@@ -480,8 +531,18 @@ enum e_optcode {
    OPT_OPENSSL_EGD,
    OPT_OPENSSL_FIPS,
    OPT_OPENSSL_KEY,
+#if HAVE_SSL_CTX_set_tlsext_max_fragment_length || defined(SSL_CTX_set_tlsext_max_fragment_length)
+   OPT_OPENSSL_MAXFRAGLEN,
+#endif
+#if HAVE_SSL_CTX_set_max_send_fragment || defined(SSL_CTX_set_max_send_fragment)
+   OPT_OPENSSL_MAXSENDFRAG,
+#endif
+   OPT_OPENSSL_MAX_PROTO_VERSION,
    OPT_OPENSSL_METHOD,
+   OPT_OPENSSL_MIN_PROTO_VERSION,
+   OPT_OPENSSL_NO_SNI,
    OPT_OPENSSL_PSEUDO,
+   OPT_OPENSSL_SNIHOST,
    OPT_OPENSSL_VERIFY,
    OPT_OPOST,		/* termios.c_oflag */
    OPT_OSPEED,		/* termios.c_ospeed */
@@ -557,6 +618,7 @@ enum e_optcode {
    OPT_PROTOCOL_FAMILY,	/* 1=PF_UNIX, 2=PF_INET, 10=PF_INET6 */
    OPT_PROXYPORT,
    OPT_PROXY_AUTHORIZATION,
+   OPT_PROXY_AUTHORIZATION_FILE,
    OPT_PROXY_RESOLVE,
 #if HAVE_DEV_PTMX || HAVE_DEV_PTC
    OPT_PTMX,
@@ -567,15 +629,20 @@ enum e_optcode {
    OPT_RANGE,		/* restrict client socket address */
    OPT_RAW,		/* termios */
    OPT_READBYTES,
+   OPT_RESET_NETNS, 	/* reset net namespace - not an option, just op! */
    OPT_RES_AAONLY,	/* resolver(3) */
    OPT_RES_DEBUG,	/* resolver(3) */
    OPT_RES_DEFNAMES,	/* resolver(3) */
    OPT_RES_DNSRCH,	/* resolver(3) */
    OPT_RES_IGNTC,	/* resolver(3) */
+   OPT_RES_NSADDR,	/* undocumented */
    OPT_RES_PRIMARY,	/* resolver(3) */
    OPT_RES_RECURSE,	/* resolver(3) */
+   OPT_RES_RETRANS,	/* undocumented */
+   OPT_RES_RETRY,	/* undocumented */
    OPT_RES_STAYOPEN,	/* resolver(3) */
    OPT_RES_USEVC,	/* resolver(3) */
+   OPT_RETRIEVE_VLAN, 	/* Linux: get VLAN info on raw sockets per auxdata */
    OPT_RETRY,
    OPT_SANE,		/* termios */
    OPT_SCTP_MAXSEG,
@@ -593,9 +660,12 @@ enum e_optcode {
    OPT_SETSID,
    OPT_SETSOCKOPT_BIN,
    OPT_SETSOCKOPT_INT,
+   OPT_SETSOCKOPT_LISTEN,
    OPT_SETSOCKOPT_STRING,
    OPT_SETUID,
    OPT_SETUID_EARLY,
+   OPT_SET_NETNS, 	/* set net namespace */
+   OPT_SHELL,
    OPT_SHUT_CLOSE,
    OPT_SHUT_DOWN,
    OPT_SHUT_NONE,
@@ -606,6 +676,7 @@ enum e_optcode {
 #ifdef SIOCSPGRP
    OPT_SIOCSPGRP,
 #endif
+   OPT_SITOUT_EIO,
 #ifdef SO_ACCEPTCONN
    OPT_SO_ACCEPTCONN,
 #endif /* SO_ACCEPTCONN */
@@ -784,17 +855,23 @@ enum e_optcode {
 #ifdef TCP_WINDOW_CLAMP
    OPT_TCP_WINDOW_CLAMP,	/* Linux 2.4.0 */
 #endif
+   OPT_TERMIOS_CFMAKERAW,	/* termios.cfmakeraw() */
+   OPT_TERMIOS_RAWER,
    OPT_TIOCSCTTY,
    OPT_TOSTOP,		/* termios.c_lflag */
    OPT_TUN_DEVICE,	/* tun: /dev/net/tun ... */
    OPT_TUN_NAME,	/* tun: tun0 */
    OPT_TUN_TYPE,	/* tun: tun|tap */
+   OPT_UDPLITE_RECV_CSCOV,
+   OPT_UDPLITE_SEND_CSCOV,
    OPT_UMASK,
-   OPT_UNIX_TIGHTSOCKLEN,	/* UNIX domain sockets */
+   OPT_UNIX_BIND_TEMPNAME,	/* UNIX domain sockets */
+   OPT_UNIX_TIGHTSOCKLEN,
    OPT_UNLINK,
    OPT_UNLINK_CLOSE,
    OPT_UNLINK_EARLY,
    OPT_UNLINK_LATE,
+   OPT_UN_UMASK,
    OPT_USER,
    OPT_USER_EARLY,
    OPT_USER_LATE,
@@ -843,38 +920,50 @@ enum e_optcode {
 enum e_phase {
    PH_ALL,		/* not for option definitions; use in apply funcs to
 			   say "all phases" */
+   PH_OFFSET, 		/* automatically applied to xio-fd */
    PH_INIT,		/* retrieving info from original state */
    PH_EARLY,		/* before any other processing */
+
    PH_PREOPEN,		/* before file descriptor is created/opened */
    PH_OPEN,		/* during filesystem entry creation/open */
    PH_PASTOPEN,		/* past filesystem entry creation/open */
+
    PH_PRESOCKET,	/* before socket call */
    PH_SOCKET,		/* for socket call */
    PH_PASTSOCKET,	/* after socket call */
+
    PH_PREBIGEN,		/* before socketpair() pipe() openpty() */
    PH_BIGEN,		/* during socketpair() pipe() openpty() */
    PH_PASTBIGEN,	/* past socketpair() pipe() openpty() */
    PH_FD,		/* soon after FD creation or identification */
+
    PH_PREBIND,		/* before socket bind() */
    PH_BIND,		/* during socket bind() ? */
    PH_PASTBIND,		/* past socket bind() - for client and server sockets! */
+
    PH_PRELISTEN,	/* before socket listen() */
    PH_LISTEN,		/* during socket listen() ? */
    PH_PASTLISTEN,	/* after socket listen() */
+
    PH_PRECONNECT,	/* before socket connect() */
    PH_CONNECT,		/* during socket connect() ? */
    PH_PASTCONNECT,	/* after socket connect() */
+
    PH_PREACCEPT,	/* before socket accept() */
    PH_ACCEPT,		/* during socket accept() ? */
    PH_PASTACCEPT,	/* after socket accept() */
    PH_CONNECTED,	/* for sockets, after connect() or accept() */
+
    PH_PREFORK,		/* before fork() (with both listen and exec!) */
    PH_FORK,		/* during fork() (with both listen and exec!) */
    PH_PASTFORK,		/* after fork() (with both listen and exec!) */
+
    PH_LATE,		/* FD is ready, before start of data loop */
    PH_LATE2,		/* FD is ready, dropping privileges */
+
    PH_PREEXEC,		/* before exec() or system() */
    PH_EXEC,		/* during exec() or system() */
+   PH_PASTEXEC,		/* only reached on addresses that do NOT exec() */
    PH_SPEC		/* specific to situation, not fix */
 } ;
 
@@ -884,7 +973,7 @@ struct optdesc {
    const char *defname;		/* default name */
    const char *nickname;	/* usual name */
    enum e_optcode optcode;	/* short form of option name */
-   unsigned int group;
+   groups_t group;
    enum e_phase phase;		/* when this option is to be used */
    enum e_types type;	/* the data type as expected on input, and stored */
    enum e_func  func;	/* which function can apply this option, e.g. ioctl(),
@@ -903,44 +992,41 @@ extern int retropt_bool(struct opt *opts, int optcode, bool *result);
 extern int retropt_short(struct opt *opts, int optcode, short *result);
 extern int retropt_ushort(struct opt *opts, int optcode, unsigned short *result);
 extern int retropt_int(struct opt *opts, int optcode, int *result);
+extern int retropt_2integrals(struct opt *opts, int optcode, union integral *value1, union integral *value2);
 extern int retropt_uint(struct opt *opts, int optcode, unsigned int *result);
 extern int retropt_long(struct opt *opts, int optcode, long *result);
 extern int retropt_ulong(struct opt *opts, int optcode, unsigned long *result);
 extern int retropt_flag(struct opt *opts, int optcode, flags_t *result);
 extern int retropt_string(struct opt *opts, int optcode, char **result);
 extern int retropt_timespec(struct opt *opts, int optcode, struct timespec *result);
-extern int retropt_bind(struct opt *opts,
-		 int af,
-		 int socktype,
-		 int ipproto,
-		 struct sockaddr *sa,
-		 socklen_t *salen,
-		 int feats,	/* TCP etc: 1..address allowed,
-				   3..address and port allowed */
-		 unsigned long res_opts0, unsigned long res_opts1);
-extern int applyopts(int fd, struct opt *opts, enum e_phase phase);
-extern int applyopts2(int fd, struct opt *opts, unsigned int from,
-		      unsigned int to);
-extern int applyopts_flags(struct opt *opts, int group, flags_t *result);
+extern int retropt_bind(struct opt *opts, int af, int socktype, int ipproto, struct sockaddr *sa, socklen_t *salen, int feats, const int ai_flags[2]);
+extern int applyopt_fd(int fd, struct opt *opt);
+extern int applyopt_single(struct single *sfd, struct opt *opt);
+extern int applyopts(struct single *sfd, int fd, struct opt *opts, enum e_phase phase);
+extern int applyopts2(struct single *sfd, int fd, struct opt *opts, unsigned int from, unsigned int to);
+extern int applyopts_optgroup(struct single *sfd, int fd, struct opt *opts, groups_t groups);
+extern int applyopts_flags(struct opt *opts, groups_t group, flags_t *result);
 extern int applyopts_cloexec(int fd, struct opt *opts);
-extern int applyopts_early(const char *path, struct opt *opts);
 extern int applyopts_fchown(int fd, struct opt *opts);
 extern int applyopts_single(struct single *fd, struct opt *opts, enum e_phase phase);
 extern int applyopts_offset(struct single *xfd, struct opt *opts);
 extern int applyopts_signal(struct single *xfd, struct opt *opts);
 extern int _xio_openlate(struct single *fd, struct opt *opts);
-extern int parseopts(const char **a, unsigned int groups, struct opt **opts);
-extern int parseopts_table(const char **a, unsigned int groups,
+extern int parseopts(const char **a, groups_t groups, struct opt **opts);
+extern int parseopts_table(const char **a, groups_t groups,
 			   struct opt **opts,
 			 const struct optname optionnames[], size_t optionnum);
-extern struct opt *copyopts(const struct opt *opts, unsigned int groups);
-extern struct opt *moveopts(struct opt *opts, unsigned int groups);
+extern const struct opt *searchopt(const struct opt *opts, groups_t groups, enum e_phase from, enum e_phase to,
+				   enum e_func func);
+extern struct opt *copyopts(const struct opt *opts, groups_t groups);
+extern struct opt *moveopts(struct opt *opts, groups_t groups);
 extern int leftopts(const struct opt *opts);
 extern int showleft(const struct opt *opts);
-extern int groupbits(int fd);
-extern int _groupbits(mode_t mode);
+extern groups_t groupbits(int fd);
+extern groups_t _groupbits(mode_t mode);
 extern int dropopts(struct opt *opts, unsigned int phase);
 extern int dropopts2(struct opt *opts, unsigned int from, unsigned int to);
+extern int dumpopts(struct opt *opts);
 
 #if HAVE_BASIC_UID_T==1
 #  define retropt_uid(o,c,r) retropt_short(o,c,r)
